"""
Everything related to meshes appropriate for the multigrid solver.
"""
# Copyright 2018-2021 The emg3d Developers.
#
# This file is part of emg3d.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.  You may obtain a copy
# of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.

import warnings
from copy import deepcopy

import numpy as np
from scipy import optimize
from scipy.constants import mu_0

from emg3d import maps

try:
    import discretize
    # Backwards compatibility; remove latest for version 1.0.0.
    dv = discretize.__version__.split('.')
    if int(dv[0]) == 0 and int(dv[1]) < 6:
        # Print is always shown and simpler, warn for the CLI logs.
        msg = ("`emg3d>=v0.15.0` ONLY works with `discretize>=v0.6.0`; "
               "`emg3d.TensorMesh` will return a basic mesh, not a "
               "`discretize.TensorMesh`.")
        print(f"* WARNING :: {msg}")
        warnings.warn(msg, UserWarning)
        dv = None
except ImportError:
    dv = None
finally:
    if dv is None:
        class dTensorMesh:
            pass
    else:
        from discretize import TensorMesh as dTensorMesh

__all__ = ['TensorMesh', 'construct_mesh', 'get_origin_widths', 'skin_depth',
           'wavelength', 'good_mg_cell_nr', 'min_cell_width',
           # Deprecated ones:
           'get_hx_h0', 'get_cell_numbers', 'get_stretched_h', 'get_domain',
           'get_hx']

MESHWARNING = (
    "`get_hx_h0`, `get_stretched_h`, `get_domain`, and `get_hx` are "
    "deprecated and will be removed. Use `construct_mesh`` instead."
)
CELLWARNING = (
    "`get_cell_numbers` is deprecated and will be removed. "
    "Use `good_mg_cell_nr` instead."
)


class _TensorMesh:
    """Minimal TensorMesh for internal multigrid computation.


    Parameters
    ----------
    h : list of three ndarrays
        Cell widths in [x, y, z] directions.

    origin : ndarray of dimension (3, )
        Origin (x, y, z).
        (For backwards compatibility one can also provide ``x0`` instead of
        ``origin``, but this will be removed in the future.)

    """

    def __init__(self, h, origin=None, **kwargs):
        """Initialize the mesh."""

        # Store origin.
        if origin is None:  # Backwards compatibility.
            origin = kwargs.pop('x0')
        self.origin = np.array(origin)

        # Width of cells, cast to arrays.
        self.h = [np.array(h[0]), np.array(h[1]), np.array(h[2])]

        # Node related properties.
        shape_nodes = (self.h[0].size+1, self.h[1].size+1, self.h[2].size+1)
        self.shape_nodes = shape_nodes
        self.n_nodes = np.prod(shape_nodes)
        self.nodes_x = np.r_[0., self.h[0].cumsum()] + self.origin[0]
        self.nodes_y = np.r_[0., self.h[1].cumsum()] + self.origin[1]
        self.nodes_z = np.r_[0., self.h[2].cumsum()] + self.origin[2]

        # Cell related properties.
        shape_cells = (self.h[0].size, self.h[1].size, self.h[2].size)
        self.shape_cells = shape_cells
        self.n_cells = np.prod(shape_cells)
        self.cell_centers_x = (self.nodes_x[1:] + self.nodes_x[:-1])/2
        self.cell_centers_y = (self.nodes_y[1:] + self.nodes_y[:-1])/2
        self.cell_centers_z = (self.nodes_z[1:] + self.nodes_z[:-1])/2

        # Edge related properties.
        self.shape_edges_x = (shape_cells[0], shape_nodes[1], shape_nodes[2])
        self.shape_edges_y = (shape_nodes[0], shape_cells[1], shape_nodes[2])
        self.shape_edges_z = (shape_nodes[0], shape_nodes[1], shape_cells[2])
        self.n_edges_x = np.prod(self.shape_edges_x)
        self.n_edges_y = np.prod(self.shape_edges_y)
        self.n_edges_z = np.prod(self.shape_edges_z)
        self.n_edges_per_direction = (
            self.n_edges_x, self.n_edges_y, self.n_edges_z)
        self.n_edges = np.sum(self.n_edges_per_direction)

        # Aliases (in line with `discretize`).
        self.x0 = self.origin
        self.nC = self.n_cells
        self.vnC = self.shape_cells
        self.nN = self.n_nodes
        self.vnN = self.shape_nodes
        self.nE = self.n_edges
        self.nEx = self.n_edges_x
        self.nEy = self.n_edges_y
        self.nEz = self.n_edges_z
        self.vnE = self.n_edges_per_direction
        self.vnEx = self.shape_edges_x
        self.vnEy = self.shape_edges_y
        self.vnEz = self.shape_edges_z

    def __repr__(self):
        """Simple representation."""
        return (f"TensorMesh: {self.shape_cells[0]} x {self.shape_cells[1]} x "
                f"{self.shape_cells[2]} ({self.n_cells:,})")

    @property
    def cell_volumes(self):
        """Construct cell volumes of the 3D model as 1D array."""
        if getattr(self, '_cell_volumes', None) is None:
            self._cell_volumes = (
                    self.h[0][None, None, :]*self.h[1][None, :, None] *
                    self.h[2][:, None, None]).ravel()
        return self._cell_volumes

    # # DEPRECATED FUNCTIONS - Remove in v0.17.0 # #
    def _warn(new_name, old_name, index=None):
        """Temporary deprecation function."""

        def get_dep(self):
            cname = type(self).__name__
            ind = f"[{index}]" if index is not None else ""
            message = (
                f"`{cname}.{old_name}` has been deprecated (will be removed "
                f"in v0.17.0); please use `{cname}.{new_name}{ind}`."
            )
            warnings.warn(message, FutureWarning)
            if index is None:
                return getattr(self, new_name)
            else:
                return getattr(self, new_name)[index]

        doc = f"`{old_name}` has been deprecated, use `{new_name}`."

        return property(get_dep, None, None, doc)

    hx = _warn('h', 'hx', 0)
    hy = _warn('h', 'hy', 1)
    hz = _warn('h', 'hz', 2)
    nCx = _warn('shape_cells', 'nCx', 0)
    nCy = _warn('shape_cells', 'nCy', 1)
    nCz = _warn('shape_cells', 'nCz', 2)
    nNx = _warn('shape_nodes', 'nNx', 0)
    nNy = _warn('shape_nodes', 'nNy', 1)
    nNz = _warn('shape_nodes', 'nNz', 2)
    vectorNx = _warn('nodes_x', 'vectorNx')
    vectorNy = _warn('nodes_y', 'vectorNy')
    vectorNz = _warn('nodes_z', 'vectorNz')
    vectorCCx = _warn('cell_centers_x', 'vectorCCx')
    vectorCCy = _warn('cell_centers_y', 'vectorCCy')
    vectorCCz = _warn('cell_centers_z', 'vectorCCz')
    vol = _warn('cell_volumes', 'vol')


class TensorMesh(dTensorMesh, _TensorMesh):
    """A slightly modified :class:`discretize.TensorMesh`.

    Adds a few attributes (`__eq__`, `copy`, `to_dict`, and `from_dict`) to the
    :class:`discretize.TensorMesh`.

    Falls back to a minimal TensorMesh if `discretize` is not installed.
    Nothing fancy is possible with the minimal TensorMesh, particularly NO
    plotting nor nice repr-functions.


    Parameters
    ----------
    h : list of three ndarrays
        Cell widths in [x, y, z] directions.

    origin : ndarray of dimension (3, )
        Origin (x, y, z).
        (For backwards compatibility one can also provide ``x0`` instead of
        ``origin``, but this will be removed in the future.)

    """

    def __eq__(self, mesh):
        """Compare two meshes.

        The provided `mesh` can be either a `emg3d` or a `discretize`
        TensorMesh.

        """

        # Check if mesh is of the same instance.
        equal = mesh.__class__.__name__ == self.__class__.__name__

        # Check dimensions.
        if equal:
            equal *= len(mesh.shape_cells) == len(self.shape_cells)

        # Check shape.
        if equal:
            equal *= np.all(self.shape_cells == mesh.shape_cells)

        # Check distances and origin.
        if equal:
            equal *= np.allclose(self.h[0], mesh.h[0], atol=0)
            equal *= np.allclose(self.h[1], mesh.h[1], atol=0)
            equal *= np.allclose(self.h[2], mesh.h[2], atol=0)
            equal *= np.allclose(self.origin, mesh.origin, atol=0)

        return bool(equal)

    def copy(self):
        """Return a copy of the TensorMesh."""
        return self.from_dict(self.to_dict(True))

    def to_dict(self, copy=False):
        """Store the necessary information of the TensorMesh in a dict."""
        out = {'hx': self.h[0], 'hy': self.h[1], 'hz': self.h[2],
               'origin': self.origin, '__class__': self.__class__.__name__}
        if copy:
            return deepcopy(out)
        else:
            return out

    @classmethod
    def from_dict(cls, inp):
        """Convert dictionary into :class:`TensorMesh` instance.

        Parameters
        ----------
        inp : dict
            Dictionary as obtained from :func:`TensorMesh.to_dict`.
            The dictionary needs the keys `hx`, `hy`, `hz`, and `origin`.

        Returns
        -------
        obj : :class:`TensorMesh` instance

        """
        try:
            # Backwards compatibility
            if 'x0' in inp.keys():
                origin = inp['x0']
            else:
                origin = inp['origin']
            return cls(h=[inp['hx'], inp['hy'], inp['hz']], origin=origin)
        except KeyError as e:
            raise KeyError(f"Variable {e} missing in `inp`.") from e


def construct_mesh(frequency, properties, center, domain=None, vector=None,
                   seasurface=None, **kwargs):
    r"""Return a TensorMesh for given parameters.

    The constructed mesh is frequency- and property-dependent. Some details are
    explained in other functions:

    - The minimum cell width :math:`\Delta_\text{min}` is a function of
      ``frequency``, ``properties[0]``, ``min_width_pps``, and
      ``min_width_limits``, see Equation :eq:`mincellwidth`.
    - The skin depth :math:`\delta` is a function of ``frequency`` and
      ``properties``, see Equation :eq:`skindepth`.
    - The wavelength :math:`\lambda` is a function of ``frequency`` and
      ``properties``, see Equation :eq:`wavelength`.

    The relation of the survey domain, computational domain, and buffer zone is
    shown in  :numref:`Figure %s <AutoGrid>` for a x-z-section; the y-direction
    behaves the same as the x-direction (the figures are only visible in the
    web version on https://emg3d.rtfd.io).

    .. figure:: ../_static/construct_mesh.png
       :align: center
       :alt: Sketch for automatic gridding.
       :name: AutoGrid

       Relation between survey domain (Ds, ``domain``), computational domain
       (Dc), and buffer zone. The survey domain should contain all sources and
       receivers as well as any important feature that should be represented in
       the data. The buffer zone is calculated as a function of wavelength with
       the provided property in the given direction.

    By default, the buffer zone around the survey domain is one wavelength.
    This means that the signal has to travel two wavelengths to get from the
    end of the survey domain to the end of the computational domain and back.
    This approach is quite conservative and on the safe side. You can reduce
    the buffer thickness if you know what you are doing. There are three
    parameters which influence the thickness of the buffer for a given
    frequency: ``properties``, which is used to calculate the skin depth and
    the wavelength, ``lambda_factor`` (default is 1) which sets how many times
    the wavelength is the thickness of the buffer (relative factor), and
    ``max_buffer``, which is an absolute maximum for the buffer thickness. A
    graphical illustration is given in :numref:`Figure %s <Buffer>`.

    .. figure:: ../_static/construct_mesh2.png
       :align: center
       :alt: Sketch for the buffer zone.
       :name: Buffer

       The thickness of the buffer zone (B) for (I)
       ``lambda_from_center=False`` (default) and for (II)
       ``lambda_from_center=True``. The ``lambda_factor``
       (:math:`\lambda_{fact}`) is a simple scaling factor for the wavelength
       :math:`\lambda`. The ``max_buffer`` is an absolute limitation.


    Parameters
    ----------

    frequency : float
        Frequency (Hz) to calculate skin depth; both the minimum cell width and
        the extent of the buffer zone, and therefore of the computational
        domain, are a function of skin depth.

    properties : float or list
        Properties to calculate the skin depths, which in turn are used to
        calculate the minimum cell width (usually at source location) and the
        extent of the buffer around the survey domain. The properties can be
        either resistivities, conductivities, or the logarithm (natural or base
        10) thereof. By default it assumes resistivities, but it can be changed
        with the parameter ``mapping``.

        Five formats are recognized for properties: it can either be a float or
        a list of 2, 3, 4, or 7 floats. Depending on the format these
        properties are used to calculate the following parameters:

        - ``p``: min_width and buffer in all directions;

        - ``[p1, p2]``:

          - 1 : min_width,
          - 2 : buffer in all directions;

        - ``[p1, p2, p3]``:

          - 1 : min_width,
          - 2 : buffer in negative z-direction,
          - 3 : buffer in all other directions;

        - ``[p1, p2, p3, p4]``:

          - 1 : min_width,
          - 2 : buffer in horizontal directions,
          - 3; 4 : buffer in negative; positive z-direction.

        - ``[p1, p2, p3, p4, p5, p6, p7]``:

          - 1 : min_width,
          - 2; 3 : buffer in negative; positive x-direction,
          - 4; 5 : buffer in negative; positive y-direction,
          - 6; 7 : buffer in negative; positive z-direction.

    center : tuple
        Tuple (or list, ndarray) of three floats for (x, y, z). The mesh is
        centered around this point, which means that here is the smallest cell.
        Usually this is the source location.

    domain : tuple of lists, list, or None, optional
        Contains the survey-domain limits. This domain should include all
        source and receiver positions as well as any important feature of the
        model. Format: ``([xmin, xmax], [ymin, ymax], [zmin, zmax])``.

        It can be None, or individual lists can be None (e.g., ``(None, None,
        [zmin, zmax])``), in which case you have to provide a ``vector``, which
        is then assumed to span exactly the domain. If only one list is
        provided it is applied to all dimensions.

    distance : tuple of lists, list, or None, optional
        An alternative to ``domain``: Instead of defining the domain in
        absolute values, they are defined here as distance from the center.
        Format: ``([xl, xr], [yl, yr], [zd, zu])``. From this the domain is
        given as ``([cx-xl, cx+xr], [cy-yl, cy+yr], [cz-zd, cz+zu])``, where
        ``center=(cx, cy, cz)``.

    vector : tuple of three ndarrays, ndarray, or None, optional
        Contains vectors of mesh-edges that should be used. If provided, the
        vector MUST at least include all of the survey domain. If ``domain``
        is not provided, it is defined as the minimum/maximum of the provided
        vector. Format: ``(xvector, yvector, zvector)``.

        It can be None, or individual ndarrays can be None (e.g., ``(xvector,
        yvector, None)``), in which case you have to provide a ``domain``. If
        only one ndarray is provided it is applied to all dimensions.

    seasurface : float, optional
        Air-sea interface. This has only to be set in the marine case, when the
        mesh in z-direction is sought for (and the interface is not contained
        in ``vector``). If set, it will ensure that at the sea surface is an
        actual boundary. It has to be bigger then the lower limit of the survey
        domain.
        Default is None.

    stretching : list or tuple of lists, optional
        Maximum stretching factors in the form of ``[max Ds, max Dc]``: the
        first value is the maximum stretching for the survey domain (default is
        1.0), the second value is the maximum stretching for the buffer zone
        (default is 1.5). If a list is provided the same is used for all three
        dimension. Alternatively a tuple of three lists can be provided, ``(x,
        y, z)``. Note that the first value has no influence on dimensions where
        a ``vector`` is provided.

    min_width_limits : float, list or None, optional
        Passed through :func:`min_cell_width` as ``limits``.
        A tuple of three can be provided for direction dependent values.
        Note that this value has no influence on dimensions where a ``vector``
        is provided.

        Default is None.

    min_width_pps : float or int, optional
        Passed through :func:`min_cell_width` as ``pps``.
        A tuple of three can be provided for direction dependent values.
        Note that this value has no influence on dimensions where a ``vector``
        is provided.

        Default is 3.

    lambda_factor : float, optional
        The buffer is taken as one wavelength from the survey domain. This can
        be regarded as quite conservative (but safe). The parameter
        ``lambda_factor`` can be used to reduce (or increase) this factor.
        Default is 1.0.

    max_buffer : float, optional
        Maximum thickness of the buffer zone around survey domain. If
        ``lambda_from_center=True``, this is the maximum distance from the
        center to the end of the computational domain.
        Default is 100,000 (100 km).

    lambda_from_center : bool, optional
        Flag how to compute the extent of the computational mesh as a function
        of wavelength:

        - False (default): The distance from the edge of the survey domain to
          the edge of the computational domain is one wavelength.
        - True: The distance from the center to the edge of the computational
          domain and back to the end of the survey domain is two wavelengths.

    mapping : str or map, optional
        Defines what type the input ``property_{x;y;z}``-values correspond to.
        By default, they represent resistivities (Ohm.m). The implemented
        mappings are:

        - 'Conductivity'; σ (S/m),
        - 'LgConductivity'; log_10(σ),
        - 'LnConductivity'; log_e(σ),
        - 'Resistivity'; ρ (Ohm.m); Default,
        - 'LgResistivity'; log_10(ρ),
        - 'LnResistivity'; log_e(ρ).

    cell_numbers : list, optional
        List of possible numbers of cells. See :func:`good_mg_cell_nr`.
        Default is ``good_mg_cell_nr(1024, 5, 3)``, which corresponds to
        numbers 16, 24, 32, 40, 48, 64, 80, 96, 128, 160, 192, 256, 320, 384,
        512, 640, 768, 1024.

    verb : int
        If 1 verbose, if 0 silent (default).
        Either way, the info is added to the returned mesh as
        ``mesh.construct_mesh_info``.


    Returns
    -------
    grid : :class:`emg3d.meshes.TensorMesh`
        Resulting mesh.


    """
    verb = kwargs.get('verb', 0)
    distance = kwargs.pop('distance', None)

    # Initiate direction-specific dicts, add unambiguous args.
    kwargs['frequency'] = frequency
    kwargs['verb'] = -1            # Run x/y/z first, print/collect
    kwargs['raise_error'] = False  # info, then raise if necessary.
    xparams = {'center': center[0]}
    yparams = {'center': center[1]}
    zparams = {'center': center[2], 'seasurface': seasurface}

    # Add properties.
    if isinstance(properties, (int, float)):
        properties = np.array([properties])
    if len(properties) == 3:
        xparams['properties'] = [properties[0], properties[2], properties[2]]
        yparams['properties'] = [properties[0], properties[2], properties[2]]
        zparams['properties'] = [properties[0], properties[1], properties[2]]
    elif len(properties) == 4:
        xparams['properties'] = [properties[0], properties[1], properties[1]]
        yparams['properties'] = [properties[0], properties[1], properties[1]]
        zparams['properties'] = [properties[0], properties[2], properties[3]]
    elif len(properties) == 7:
        xparams['properties'] = [properties[0], properties[1], properties[2]]
        yparams['properties'] = [properties[0], properties[3], properties[4]]
        zparams['properties'] = [properties[0], properties[5], properties[6]]
    else:
        kwargs['properties'] = properties

    # Add optionally direction specific args.
    for name, value in zip(['domain', 'vector', 'distance'],
                           [domain, vector, distance]):
        if (value is not None and len(value) == 3 and not
                isinstance(value, np.ndarray)):
            if value[0] is not None:
                xparams[name] = value[0]
            if value[1] is not None:
                yparams[name] = value[1]
            if value[2] is not None:
                zparams[name] = value[2]
        else:
            kwargs[name] = value

    # Add optionally direction specific kwargs.
    for name in ['stretching', 'min_width_limits', 'min_width_pps']:
        value = kwargs.pop(name, None)
        if value is not None:
            if isinstance(value, (int, float)):
                kwargs[name] = np.array([value])
            elif len(value) == 3:
                if value[0] is not None:
                    xparams[name] = value[0]
                if value[1] is not None:
                    yparams[name] = value[1]
                if value[2] is not None:
                    zparams[name] = value[2]
            else:
                kwargs[name] = value

    # Get origins and widths in all directions.
    x0, hx, xinfo = get_origin_widths(**kwargs, **xparams)
    y0, hy, yinfo = get_origin_widths(**kwargs, **yparams)
    z0, hz, zinfo = get_origin_widths(**kwargs, **zparams)

    # Throw message if no solution was found.
    if any([out is None for out in [x0, y0, z0]]):
        raise RuntimeError("No suitable grid found; relax your criteria.")

    # Create mesh.
    mesh = TensorMesh([hx, hy, hz], origin=np.array([x0, y0, z0]))

    # Collect info.
    info = (f"\n         == GRIDDING IN X ==\n{xinfo}\n"
            f"\n         == GRIDDING IN Y ==\n{yinfo}\n"
            f"\n         == GRIDDING IN Z ==\n{zinfo}\n")
    mesh.construct_mesh_info = info
    if verb > 0:
        print(info)

    return mesh


def get_origin_widths(frequency, properties, center, domain=None, vector=None,
                      seasurface=None, **kwargs):
    r"""Return origin and cell widths for given parameters.

    This function works in one dimension only, and is used by
    :func:`construct_mesh` once in each direction. It is recommended to use
    directly function :func:`construct_mesh`, which returns a
    :class:`TensorMesh`.

    All the parameters are described in :func:`construct_mesh`. Described here
    are only the differences.


    Parameters
    ----------

    All : description
        All parameters are described in :func:`construct_mesh`. The only
        difference is that here only variables for one direction are accepted.

    verb : int
        If 1 verbose, if 0 silent (default); if -1 it returns the info as
        string instead of printing it.

    raise_error : bool, optional
        If True, an error is raised if no suitable grid is found. Otherwise it
        just returns None's (used by ``construct_mesh``).
        Default is True.

    Returns
    -------
    origin : float
        Origin of the mesh.

    widths : ndarray
        Cell widths of mesh.

    info : str
        Info, only if ``verb<0`.

    """
    # Get all kwargs.
    distance = kwargs.pop('distance', None)
    stretching = kwargs.pop('stretching', [1.0, 1.5])
    min_width_limits = kwargs.pop('min_width_limits', None)
    min_width_pps = kwargs.pop('min_width_pps', 3)
    lambda_factor = kwargs.pop('lambda_factor', 1.0)
    max_buffer = kwargs.pop('max_buffer', 100000)
    lambda_from_center = kwargs.pop('lambda_from_center', False)
    pmap = kwargs.pop('mapping', 'Resistivity')
    cell_numbers = kwargs.pop('cell_numbers', good_mg_cell_nr())
    raise_error = kwargs.pop('raise_error', True)
    verb = kwargs.pop('verb', 0)

    # Ensure no kwargs left.
    if kwargs:
        raise TypeError(f"Unexpected **kwargs: {list(kwargs.keys())}")

    # Get property map from string.
    if isinstance(pmap, str):
        pmap = getattr(maps, 'Map'+pmap)()

    # Properties.
    cond = pmap.backward(np.array(properties, ndmin=1, dtype=float))
    cond_arr = np.array([
        cond[0], cond[min(cond.size-1, 1)], cond[min(cond.size-1, 2)]])

    # Get skin depth.
    skind = skin_depth(frequency, cond_arr)

    # Minimum cell width.
    dmin = min_cell_width(skind[0], min_width_pps, min_width_limits)

    # Survey domain: if not provided get from vector or distance.
    # Priority: domain > vector > distance.
    if domain is None and vector is None and distance is None:
        raise ValueError(
                "At least one of `domain`, `distance`, and `vector` "
                "must be provided.")
    elif domain is None:
        if vector is None:
            domain = np.array([center-abs(distance[0]),
                               center+abs(distance[1])])
        else:
            domain = np.array([vector.min(), vector.max()], dtype=float)
    else:
        domain = np.array(domain, dtype=np.float64)
        if vector is not None:
            if domain[0] < vector.min() or domain[1] > vector.max():
                raise ValueError(
                        "Provided vector MUST at least include all of the "
                        "survey domain.")

    # Seasurface related checks.
    if seasurface is not None:

        # Check that seasurface > center.
        if seasurface <= center:
            raise ValueError(
                    "The `seasurface` but be bigger then `center`.")

        # If center is close to seasurface, set it to seasurface.
        if abs(seasurface - center) < dmin:
            center = seasurface

    # Computation domain; big enough to avoid boundary effects.
    # To avoid boundary effects we want the signal to travel two wavelengths
    # from the source to the boundary and back to the receiver.
    # => 2*pi*sd ~ 6.3*sd = one wavelength => signal is ~ 0.2 %.
    # Two wavelengths we can safely assume it is zero.
    wlength = lambda_factor*wavelength(skind[1:])

    if lambda_from_center:
        # Center to edges of domain.
        in_domain = abs(domain - center)

        # Required buffer, additional to domain.
        d_buff = np.max([np.zeros(2), (2*wlength - in_domain)/2], axis=0)

        # Add buffer to domain.
        comp_domain = np.array([domain[0]-d_buff[0], domain[1]+d_buff[1]])

        # Restrict total domain to max_buffer.
        comp_domain[0] = max(comp_domain[0], center-max_buffer)
        comp_domain[1] = min(comp_domain[1], center+max_buffer)

    else:
        dbuffer = np.min([wlength, np.ones(2)*max_buffer], axis=0)
        comp_domain = np.array([domain[0]-dbuffer[0], domain[1]+dbuffer[1]])

    # Initiate flag if terminated.
    finished = False

    # Initiate alpha variables for survey and computation domains.
    sa, ca = 1.0, 1.0

    # Loop over possible cell numbers from small to big.
    for nx in np.unique(cell_numbers):

        # Loop over possible alphas for domain.
        for sa in np.arange(1.0, stretching[0]+0.005, 0.01):

            if vector is None:

                # Get current stretched grid cell sizes.
                thxl = dmin*sa**np.arange(nx)  # Left of origin.
                thxr = dmin*sa**np.arange(nx)  # Right of origin.

                # Adjust stretching for seasurface if required.
                if seasurface is not None and seasurface > center:
                    t_nx = np.r_[center, center+np.cumsum(thxr)]
                    ii = np.argmin(abs(t_nx-seasurface))
                    thxr[:ii] *= abs(seasurface-center)/np.sum(thxr[:ii])

                # Fill from center to left and right domain.
                nl = np.sum((center-np.cumsum(thxl)) > domain[0]) + 1
                nr = np.sum((center+np.cumsum(thxr)) < domain[1]) + 1

                # Create the current hx-array.
                hx = np.r_[thxl[:nl][::-1], thxr[:nr]]

                # Get actual domain:
                asurv_domain = [center - np.sum(thxl[:nl]),
                                center + np.sum(thxr[:nr])]

            else:
                # Store actual domain, current hx-array, and number of cells.
                asurv_domain = [vector[0], vector[-1]]
                hx = np.diff(vector)

            # Expand for seasurface if necessary.
            if seasurface is not None and seasurface > asurv_domain[-1]:
                thxr = hx[-1]*sa**np.arange(nx)
                sdepth = seasurface - asurv_domain[-1]

                # Get number of element, round down, and stretch.
                ii = np.argmax(np.cumsum(thxr) > sdepth)
                thxr = thxr[:ii]  # Restrict.
                thxr *= abs(seasurface-asurv_domain[-1])/np.sum(thxr)

                # Adjust actual domain, hx, and count.
                asurv_domain[1] += sum(thxr)
                hx = np.r_[hx, thxr]

            # Remaining number of cells
            nx_remain = nx - hx.size

            # Not good, try next.
            if nx_remain <= 0:
                continue

            # Store for verbosity
            hxo = hx

            # Loop over possible alphas for buffer.
            for ca in np.arange(sa, stretching[1]+0.005, 0.01):

                # Get current stretched grid cell sizes.
                thxl = hx[0]*ca**np.arange(1, nx_remain+1)   # Left of survey.
                thxr = hx[-1]*ca**np.arange(1, nx_remain+1)  # Right of survey.

                # Fill from survey to left and right domain.
                nl = np.sum((asurv_domain[0] - np.cumsum(thxl)) >
                            comp_domain[0]) + 1
                nr = np.sum((asurv_domain[1] + np.cumsum(thxr)) <
                            comp_domain[1]) + 1

                # Get remaining number of cells.
                nx_remain2 = nx_remain - nl - nr

                if nx_remain2 < 0:  # Not good, try next.
                    continue

                # Create hx-array.
                nl += int(np.floor(nx_remain2/2))  # If uneven, add one cell
                nr += int(np.ceil(nx_remain2/2))   # more on the right.
                hx = np.r_[thxl[:nl][::-1], hx, thxr[:nr]]

                # Compute origin.
                x0 = float(asurv_domain[0]-np.sum(thxl[:nl]))

                # Mark it as finished and break out of the loop.
                finished = True
                break

            if finished:
                break

        if finished:
            break

    # Raise Error or return Nones if not finish.
    if not finished:
        msg = "No suitable grid found; relax your criteria."
        # Throw message if no solution was found.
        if raise_error:
            raise RuntimeError(msg)
        else:
            x0, hx, info = None, None, msg

    else:  # Collect info about final grid.

        # Check max stretching.
        sa_adj = np.max([hxo[1:]/hxo[:-1], hxo[:-1]/hxo[1:]])
        sa_limit = min(1.5, stretching[0]+0.25)

        # Display precision.
        prec = int(np.ceil(max(0, -np.log10(min(hx))+1)))

        info = f"Skin depth     [m] : {skind[0]:.{prec}f}"
        if cond.size > 1:
            info += f" / {skind[1]:.{prec}f}"
        if cond.size > 2:
            info += f" / {skind[2]:.{prec}f}"
        info += "  [corr. to `properties`]\n"

        info += (
            f"Survey dom. DS [m] : "
            f"{domain[0]:.{prec}f} - {domain[1]:.{prec}f}\n"
            #
            f"Comp. dom. DC  [m] : {comp_domain[0]:.{prec}f} - "
            f"{comp_domain[1]:.{prec}f}\n"
            #
            f"Final extent   [m] : {x0:.{prec}f} - {x0+np.sum(hx):.{prec}f}\n"
            #
            f"Cell widths    [m] : {min(hxo):.{prec}f} / {max(hxo):.{prec}f} "
            f"/ {max(hx):.{prec}f}  [min(DS) / max(DS) / max(DC)]\n"
            #
            f"Number of cells    : {nx} ({hxo.size} / "
            f"{nx-hxo.size-nx_remain2} / {nx_remain2})  "
            f"[Total (DS/DC/remain)]\n"
            #
            f"Max stretching     : {sa:.3f} ({sa_adj:.3f}) / {ca:.3f}"
            "  [DS (seasurface) / DC]")

        if sa_adj > sa_limit:
            info += (f"\nNote: Stretching in DS >> {sa}.\nThe reason "
                     "is usually the interplay of center/domain/seasurface.")

    if verb > 0:
        print(info)

    # Return required info.
    if verb < 0:
        return x0, hx, info
    else:
        return x0, hx


def good_mg_cell_nr(max_nr=1024, max_prime=5, min_div=3):
    r"""Returns 'good' cell numbers for the multigrid method.

    'Good' cell numbers are numbers which can be divided by 2 as many times as
    possible. At the end there will be a low prime number.

    The function adds all numbers :math:`p 2^n \leq M` for :math:`p={2, 3, ...,
    p_\text{max}}` and :math:`n={n_\text{min}, n_\text{min}+1, ..., \infty}`;
    :math:`M, p_\text{max}, n_\text{min}` correspond to `max_nr`, `max_prime`,
    and `min_div`, respectively.


    Parameters
    ----------
    max_nr : int, optional
        Maximum number of cells.
        Default is 1024.

    max_prime : int, optional
        Highest permitted prime number p for p*2^n. {2, 3, 5, 7} are good upper
        limits in order to avoid too big lowest grids in the multigrid method.
        Default is 5.

    min_div : int, optional
        Minimum times the number can be divided by two.
        Default is 3.


    Returns
    -------
    numbers : array
        Array containing all possible cell numbers from lowest to highest.

    """
    # Primes till 20.
    primes = np.array([2, 3, 5, 7, 11, 13, 17, 19], dtype=np.int64)

    # Sanity check; 19 is already ridiculously high.
    if max_prime > primes[-1]:
        raise ValueError(
                f"Highest prime is {max_prime}, please use a value < 20.")

    # Restrict to max_prime.
    primes = primes[primes <= max_prime]

    # Get possible values.
    # Currently restricted to prime*2**30 (for prime=2 => 1,073,741,824 cells).
    numbers = primes[:, None]*2**np.arange(min_div, 30)

    # Get unique values.
    numbers = np.unique(numbers)

    # Restrict to max_nr and return.
    return numbers[numbers <= max_nr]


def skin_depth(frequency, conductivity, mu=mu_0):
    r"""Return skin depth for provided frequency and conductivity.

    The skin depth :math:`\delta` (m) is given by

    .. math::
        :label: skindepth

        \delta = \sqrt{\frac{2}{\omega\sigma\mu}}\ ,

    where :math:`\omega=2\pi f` is angular frequency of frequency :math:`f`
    (Hz), :math:`\sigma` is conductivity (S/m), and :math:`\mu` is magnetic
    permeability (H/m).


    Parameters
    ----------
    frequency : float
        Frequency (Hz).

    conductivity : float
        Conductivity (S/m).

    mu : float, optional
        Magnetic permeability (H/m); default is :math:`\mu_0`.


    Returns
    -------
    skindepth : float
        Skin depth (m).

    """
    skind = 1/np.sqrt(np.pi*abs(frequency)*conductivity*mu)
    if frequency < 0:  # For Laplace-domain computations.
        skind /= np.sqrt(2*np.pi)

    return skind


def wavelength(skin_depth):
    r"""Return the wavelength.

    The wavelength :math:`\lambda` (m) is given by

    .. math::
        :label: wavelength

        \lambda = 2\pi\delta\ .

    The skin depth :math:`\delta` is a function of frequency and conductivity
    and is given by :func:`skin_depth`, Equation :eq:`skindepth`.


    Parameters
    ----------
    skin_depth : float or ndarray.
        Skin depth (m).


    Returns
    -------
    wavelength : float or ndarray
        Wavelength (m).

    """
    return 2*np.pi*skin_depth


def min_cell_width(skin_depth, pps=3, limits=None):
    r"""Return the minimum cell width.

    The minimum cell width is defined by the desired points per skin depth,

    .. math::
        :label: mincellwidth

        \Delta_\text{min} =
        \text{limits[0]} \le \frac{\delta}{\text{pps}} \le \text{limits[1]} \ .

    The skin depth :math:`\delta` is a function of frequency and conductivity
    and is given by :func:`skin_depth`, Equation :eq:`skindepth`.


    Parameters
    ----------
    skin_depth : float
        Skin depth (m).

    pps : int
        Points per skin depth.

    limits : None, float, or list of two floats
        Limits on minimum width:

        - None: No limits.
        - float: Returns limits as dmin.
        - [min, max]: dmin is limited to this range.


    Returns
    -------
    dmin : float
        Minimum cell width (m).

    """
    # Calculate min cell width.
    dmin = skin_depth/pps

    # Respect user limits.
    if limits is not None:

        limits = np.array(limits, ndmin=1)

        if limits.size == 1:
            dmin = limits  # Ignores skin depth and pps.
        else:
            dmin = np.clip(dmin, *limits)  # Restrict.

    return dmin


# # # DEPRECATED FUNCTIONS # # #
def get_hx_h0(freq, res, domain, fixed=0., possible_nx=None, min_width=None,
              pps=3, alpha=None, max_domain=100000., raise_error=True, verb=1,
              return_info=False):
    r"""Return cell widths and origin for given parameters.

    Returns cell widths for the provided frequency, resistivity, domain extent,
    and other parameters using a flexible amount of cells. See input parameters
    for more details. A maximum of three hard/fixed boundaries can be provided
    (one of which is the grid center).

    The minimum cell width is computed through :math:`\delta/\rm{pps}`, where
    the skin depth is given by :math:`\delta = 503.3 \sqrt{\rho/f}`, and the
    parameter `pps` stands for 'points-per-skindepth'. The minimum cell width
    can be restricted with the parameter `min_width`.

    The actual computation domain adds a buffer zone around the (survey)
    domain. The thickness of the buffer is six times the skin depth. The field
    is basically zero after two wavelengths. A wavelength is
    :math:`2\pi\delta`, hence roughly 6 times the skin depth. Taking a factor 6
    gives therefore almost two wavelengths, as the field travels to the
    boundary and back. The actual buffer thickness can be steered with the
    `res` parameter.

    One has to take into account that the air is very resistive, which has to
    be considered not just in the vertical direction, but also in the
    horizontal directions, as the airwave will bounce back from the sides
    otherwise. In the marine case this issue reduces with increasing water
    depth.


    See Also
    --------
    get_stretched_h : Get `hx` for a fixed number `nx` and within a fixed
                      domain.


    Parameters
    ----------

    freq : float
        Frequency (Hz) to compute the skin depth. The skin depth is a concept
        defined in the frequency domain. If a negative frequency is provided,
        it is assumed that the computation is carried out in the Laplace
        domain. To compute the skin depth, the value of `freq` is then
        multiplied by :math:`-2\pi`, to simulate the closest
        frequency-equivalent.

    res : float or list
        Resistivity (Ohm m) to compute the skin depth. The skin depth is
        used to compute the minimum cell width and the boundary thicknesses.
        Up to three resistivities can be provided:

        - float: Same resistivity for everything;
        - [min_width, boundaries];
        - [min_width, left boundary, right boundary].

    domain : list
        Contains the survey-domain limits [min, max]. The actual computation
        domain consists of this domain plus a buffer zone around it, which
        depends on frequency and resistivity.

    fixed : list, optional
        Fixed boundaries, one, two, or maximum three values. The grid is
        centered around the first value. Hence it is the center location with
        the smallest cell. Two more fixed boundaries can be added, at most one
        on each side of the first one.
        Default is 0.

    possible_nx : list, optional
        List of possible numbers of cells. See :func:`good_mg_cell_nr`.
        Default is ``good_mg_cell_nr(1024, 5, 3)``, which corresponds to
        numbers 16, 24, 32, 40, 48, 64, 80, 96, 128, 160, 192, 256, 320, 384,
        512, 640, 768, 1024.

    min_width : float, list or None, optional
        Minimum cell width restriction:

        - None : No restriction;
        - float : Fixed to this value, ignoring skin depth and `pps`.
        - list [min, max] : Lower and upper bounds.

        Default is None.

    pps : int, optional
        Points per skindepth; minimum cell width is computed via
        `dmin = skindepth/pps`.
        Default = 3.

    alpha : list, optional
        Maximum alpha and step size to find a good alpha. The first value is
        the maximum alpha of the survey domain, the second value is the maximum
        alpha for the buffer zone, and the third value is the step size.
        Default = [1, 1.5, .01], hence no stretching within the survey domain
        and a maximum stretching of 1.5 in the buffer zone; step size is 0.01.

    max_domain : float, optional
        Maximum computation domain from fixed[0] (usually source position).
        Default is 100,000.

    raise_error : bool, optional
        If True, an error is raised if no suitable grid is found. Otherwise it
        just prints a message and returns None's.
        Default is True.

    verb : int, optional
        Verbosity, 0 or 1.
        Default = 1.

    return_info : bool
        If True, a dictionary is returned with some grid info (min and max
        cell width and alpha).


    Returns
    -------
    hx : ndarray
        Cell widths of mesh.

    x0 : float
        Origin of the mesh.

    info : dict
        Dictionary with mesh info; only if ``return_info=True``.

        Keys:

        - `dmin`: Minimum cell width;
        - `dmax`: Maximum cell width;
        - `amin`: Minimum alpha;
        - `amax`: Maximum alpha.

    """
    warnings.warn(MESHWARNING, FutureWarning)

    # Get variables with default lists:
    if alpha is None:
        alpha = [1, 1.5, 0.01]
    if possible_nx is None:
        possible_nx = good_mg_cell_nr()

    # Cast resistivity value(s).
    res = np.array(res, ndmin=1)
    if res.size == 1:
        res_arr = np.array([res[0], res[0], res[0]])
    elif res.size == 2:
        res_arr = np.array([res[0], res[1], res[1]])
    else:
        res_arr = np.array([res[0], res[1], res[2]])

    # Cast and check fixed.
    fixed = np.array(fixed, ndmin=1)
    if fixed.size > 2:

        # Check length.
        if fixed.size > 3:
            raise ValueError("Maximum three fixed boundaries permitted. "
                             f"Provided: {fixed.size}.")

        # Sort second and third, so it doesn't matter how it was provided.
        fixed = np.array([fixed[0], max(fixed[1:]), min(fixed[1:])])

        # Check side.
        if np.sign(np.diff(fixed[:2])) == np.sign(np.diff(fixed[::2])):
            raise ValueError(
                    "2nd and 3rd fixed boundaries have to be left and right "
                    "of the first one.\n"
                    f"Provided: [{fixed[0]}, {fixed[1]}, {fixed[2]}]")

    # Get skin depth.
    skind = skin_depth(freq, 1/res_arr)

    # Minimum cell width.
    dmin = min_cell_width(skind[0], pps, min_width)

    # Survey domain; contains all sources and receivers.
    domain = np.array(domain, dtype=np.float64)

    # Computation domain; big enough to avoid boundary effects.
    # To avoid boundary effects we want the signal to travel two wavelengths
    # from the source to the boundary and back to the receiver.
    # => 2*pi*sd ~ 6.3*sd = one wavelength => signal is ~ 0.2 %.
    # Two wavelengths we can safely assume it is zero.
    #
    # The air does not follow the concept of skin depth, as it is a wave rather
    # than diffusion. For this is the factor `max_domain`, which restricts
    # the domain in each direction to this value from the center.

    # (a) Source to edges of domain.
    dist_in_domain = abs(domain - fixed[0])

    # (b) Two wavelengths.
    two_lambda = 2*wavelength(skind[1:])

    # (c) Required buffer, additional to domain.
    dist_buff = np.max([np.zeros(2), (two_lambda - dist_in_domain)/2], axis=0)

    # (d) Add buffer to domain.
    comp_domain = np.array([domain[0]-dist_buff[0], domain[1]+dist_buff[1]])

    # (e) Restrict total domain to max_domain.
    comp_domain[0] = max(comp_domain[0], fixed[0]-max_domain)
    comp_domain[1] = min(comp_domain[1], fixed[0]+max_domain)

    # Initiate flag if terminated.
    finished = False

    # Initiate alpha variables for survey and computation domains.
    sa, ca = 1.0, 1.0

    # Loop over possible cell numbers from small to big.
    for nx in np.unique(possible_nx):

        # Loop over possible alphas for domain.
        for sa in np.arange(1.0, alpha[0]+alpha[2]/2, alpha[2]):

            # Get current stretched grid cell sizes.
            thxl = dmin*sa**np.arange(nx)  # Left of origin.
            thxr = dmin*sa**np.arange(nx)  # Right of origin.

            # 0. Adjust stretching for fixed boundaries.
            if fixed.size > 1:  # Move mesh to first fixed boundary.
                t_nx = np.r_[fixed[0], fixed[0]+np.cumsum(thxr)]
                ii = np.argmin(abs(t_nx-fixed[1]))
                thxr *= abs(fixed[1]-fixed[0])/np.sum(thxr[:ii])

            if fixed.size > 2:  # Move mesh to second fixed boundary.
                t_nx = np.r_[fixed[0], fixed[0]-np.cumsum(thxl)]
                ii = np.argmin(abs(t_nx-fixed[2]))
                thxl *= abs(fixed[2]-fixed[0])/np.sum(thxl[:ii])

            # 1. Fill from center to left domain.
            nl = np.sum((fixed[0]-np.cumsum(thxl)) > domain[0])+1

            # 2. Fill from center to right domain.
            nr = np.sum((fixed[0]+np.cumsum(thxr)) < domain[1])+1

            # 3. Get remaining number of cells and check termination criteria.
            nsdc = nl+nr  # Number of domain cells.
            nx_remain = nx-nsdc

            # Not good, try next.
            if nx_remain <= 0:
                continue

            # Create the current hx-array.
            hx = np.r_[thxl[:nl][::-1], thxr[:nr]]
            hxo = np.r_[thxl[:nl][::-1], thxr[:nr]]

            # Get actual domain:
            asurv_domain = [fixed[0]-np.sum(thxl[:nl]),
                            fixed[0]+np.sum(thxr[:nr])]
            x0 = float(fixed[0]-np.sum(thxl[:nl]))

            # Get actual stretching (differs in case of fixed layers).
            sa_adj = np.max([hx[1:]/hx[:-1], hx[:-1]/hx[1:]])

            # Loop over possible alphas for comp_domain.
            for ca in np.arange(sa, alpha[1]+alpha[2]/2, alpha[2]):

                # 4. Fill to left comp_domain.
                thxl = hx[0]*ca**np.arange(1, nx_remain+1)
                nl = np.sum((asurv_domain[0]-np.cumsum(thxl)) >
                            comp_domain[0])+1

                # 5. Fill to right comp_domain.
                thxr = hx[-1]*ca**np.arange(1, nx_remain+1)
                nr = np.sum((asurv_domain[1]+np.cumsum(thxr)) <
                            comp_domain[1])+1

                # 6. Get remaining number of cells and check termination
                # criteria.
                ncdc = nl+nr  # Number of comp_domain cells.
                nx_remain2 = nx-nsdc-ncdc

                if nx_remain2 < 0:  # Not good, try next.
                    continue

                # Create hx-array.
                nl += int(np.floor(nx_remain2/2))  # If uneven, add one cell
                nr += int(np.ceil(nx_remain2/2))   # more on the right.
                hx = np.r_[thxl[:nl][::-1], hx, thxr[:nr]]

                # Compute origin.
                x0 = float(asurv_domain[0]-np.sum(thxl[:nl]))

                # Mark it as finished and break out of the loop.
                finished = True
                break

            if finished:
                break

        if finished:
            break

    # Check finished and print info about found grid.
    if not finished:
        # Throw message if no solution was found.
        if raise_error:
            raise RuntimeError("No suitable grid found; relax your criteria.")
        else:
            print("* ERROR   :: No suitable grid found; relax your criteria.")
            hx, x0 = None, None

    elif verb > 0:
        print("   Skin depth ", end="")
        if res.size == 1:
            print(f"         [m] : {skind[0]:.0f}")
        elif res.size == 2:
            print(f"(m/l-r)  [m] : {skind[0]:.0f} / {skind[1]:.0f}")
        else:
            print(f"(m/l/r)  [m] : {skind[0]:.0f} / {skind[1]:.0f} / "
                  f"{skind[2]:.0f}")
        print(f"   Survey domain       [m] : {domain[0]:.0f} - "
              f"{domain[1]:.0f}")
        print(f"   Computation domain  [m] : {comp_domain[0]:.0f} - "
              f"{comp_domain[1]:.0f}")
        print(f"   Final extent        [m] : {x0:.0f} - "
              f"{x0+np.sum(hx):.0f}")
        extstr = f"   Min/max cell width  [m] : {min(hx):.0f} / "
        alstr = "   Alpha survey"
        nrstr = "   Number of cells "
        if not np.isclose(sa, sa_adj):
            sastr = f"{sa:.3f} ({sa_adj:.3f})"
        else:
            sastr = f"{sa:.3f}"
        print(extstr+f"{max(hxo):.0f} / {max(hx):.0f}")
        print(alstr+f"/comp       : {sastr} / {ca:.3f}")
        print(nrstr+f"(s/c/r) : {nx} ({nsdc}/{ncdc}/{nx_remain2})")
        print()

    if return_info:
        if not fixed.size > 1:
            sa_adj = sa

        info = {'dmin': dmin,
                'dmax': np.nanmax(hx),
                'amin': np.nanmin([ca, sa, sa_adj]),
                'amax': np.nanmax([ca, sa, sa_adj])}

        return hx, x0, info
    else:
        return hx, x0


def get_cell_numbers(max_nr, max_prime=5, min_div=3):
    warnings.warn(CELLWARNING, FutureWarning)
    return good_mg_cell_nr(max_nr, max_prime, min_div)


def get_stretched_h(min_width, domain, nx, x0=0, x1=None, resp_domain=False):
    """Return cell widths for a stretched grid within the domain.

    Returns `nx` cell widths within `domain`, where the minimum cell width is
    `min_width`. The cells are not stretched within `x0` and `x1`, and outside
    uses a power-law stretching. The actual stretching factor and the number of
    cells left and right of `x0` and `x1` are found in a minimization process.

    The domain is not completely respected. The starting point of the domain
    is, but the endpoint of the domain might slightly shift (this is more
    likely the case for small `nx`, for big `nx` the shift should be small).
    The new endpoint can be obtained with ``domain[0]+np.sum(hx)``. If you want
    the domain to be respected absolutely, set ``resp_domain=True``. However,
    be aware that this will introduce one stretch-factor which is different
    from the other stretch factors, to accommodate the restriction. This
    one-off factor is between the left- and right-side of `x0`, or, if `x1` is
    provided, just after `x1`.


    See Also
    --------
    get_hx_x0 : Get `hx` and `x0` for a flexible number of `nx` with
                given bounds.


    Parameters
    ----------

    min_width : float
        Minimum cell width. If x1 is provided, the actual minimum cell width
        might be smaller than min_width.

    domain : list
        [start, end] of model domain.

    nx : int
        Number of cells.

    x0 : float
        Center of the grid. `x0` is restricted to `domain`.
        Default is 0.

    x1 : float
        If provided, then no stretching is applied between `x0` and `x1`. The
        non-stretched part starts at `x0` and stops at the first possible
        location at or after `x1`. `x1` is restricted to `domain`. This will
        min_width so that an integer number of cells fit within x0 and x1.

    resp_domain : bool
        If False (default), then the domain-end might shift slightly to assure
        that the same stretching factor is applied throughout. If set to True,
        however, the domain is respected absolutely. This will introduce one
        stretch-factor which is different from the other stretch factors, to
        accommodate the restriction. This one-off factor is between the left-
        and right-side of `x0`, or, if `x1` is provided, just after `x1`.


    Returns
    -------
    hx : ndarray
        Cell widths of mesh.

    """
    warnings.warn(MESHWARNING, FutureWarning)

    # Cast to arrays
    domain = np.array(domain, dtype=np.float64)
    x0 = np.array(x0, dtype=np.float64)
    x0 = np.clip(x0, *domain)  # Restrict to model domain
    min_width = np.array(min_width, dtype=np.float64)
    if x1 is not None:
        x1 = np.array(x1, dtype=np.float64)
        x1 = np.clip(x1, *domain)  # Restrict to model domain

    # If x1 is provided (a part is not stretched)
    if x1 is not None:

        # Store original values
        xlim_orig = domain.copy()
        nx_orig = int(nx)
        x0_orig = x0.copy()
        h_min_orig = min_width.copy()

        # Get number of non-stretched cells
        n_nos = int(np.ceil((x1-x0)/min_width))

        # Re-compute min_width to fit with x0-x1-limits:
        min_width = (x1-x0)/n_nos

        # Subtract one cell, because the standard scheme provides one
        # min_width-cell.
        n_nos -= 1

        # Reset x0, because the first min_width comes from normal scheme
        x0 += min_width

        # Reset xmax for normal scheme
        domain[1] -= n_nos*min_width

        # Reset nx for normal scheme
        nx -= n_nos

        # If there are not enough points reset to standard procedure. The limit
        # of five is arbitrary. However, nx should be much bigger than five
        # anyways, otherwise stretched grid doesn't make sense.
        if nx <= 5:
            print("Warning :: Not enough points for non-stretched part,"
                  "ignoring therefore `x1`.")
            domain = xlim_orig
            nx = nx_orig
            x0 = x0_orig
            x1 = None
            min_width = h_min_orig

    # Get stretching factor (a = 1+alpha).
    if min_width == 0 or min_width > np.diff(domain)/nx:
        # If min_width is bigger than the domain-extent divided by nx, no
        # stretching is required at all.
        alpha = 0
    else:

        # Wrap _get_dx into a minimization function to call with fsolve.
        def find_alpha(alpha, min_width, args):
            """Find alpha such that min(hx) = min_width."""
            return min(get_hx(alpha, *args))/min_width-1

        # Search for best alpha, must be at least 0
        args = (domain, nx, x0)
        alpha = max(0, optimize.fsolve(find_alpha, 0.02, (min_width, args)))

    # With alpha get actual cell spacing with `resp_domain` to respect the
    # users decision.
    hx = get_hx(alpha, domain, nx, x0, resp_domain)

    # Add the non-stretched center if x1 is provided
    if x1 is not None:
        hx = np.r_[hx[: np.argmin(hx)], np.ones(n_nos)*min_width,
                   hx[np.argmin(hx):]]

    # Print warning min_width could not be respected.
    if abs(hx.min() - min_width) > 0.1:
        print(f"Warning :: Minimum cell width ({np.round(hx.min(), 2)} m) is "
              "below `min_width`, because `nx` is too big for `domain`.")

    return hx


def get_domain(x0=0, freq=1, res=0.3, limits=None, min_width=None,
               fact_min=0.2, fact_neg=5, fact_pos=None):
    r"""Get domain extent and minimum cell width as a function of skin depth.

    Returns the extent of the computation domain and the minimum cell width as
    a multiple of the skin depth, with possible user restrictions on minimum
    computation domain and range of possible minimum cell widths.

    .. math::

            \delta &= 503.3 \sqrt{\frac{\rho}{f}} , \\
            x_\text{start} &= x_0-k_\text{neg}\delta , \\
            x_\text{end} &= x_0+k_\text{pos}\delta , \\
            h_\text{min} &= k_\text{min} \delta .


    Parameters
    ----------

    x0 : float
        Center of the computation domain. Normally the source location.
        Default is 0.

    freq : float
        Frequency (Hz) to compute the skin depth. The skin depth is a concept
        defined in the frequency domain. If a negative frequency is provided,
        it is assumed that the computation is carried out in the Laplace
        domain. To compute the skin depth, the value of `freq` is then
        multiplied by :math:`-2\pi`, to simulate the closest
        frequency-equivalent.

        Default is 1 Hz.

    res : float, optional
        Resistivity (Ohm m) to compute skin depth.
        Default is 0.3 Ohm m (sea water).

    limits : None or list
        [start, end] of model domain. This extent represents the minimum extent
        of the domain. The domain is therefore only adjusted if it has to reach
        outside of [start, end].
        Default is None.

    min_width : None, float, or list of two floats
        Minimum cell width is computed as a function of skin depth:
        fact_min*sd. If `min_width` is a float, this is used. If a list of
        two values [min, max] are provided, they are used to restrain
        min_width. Default is None.

    fact_min, fact_neg, fact_pos : floats
        The skin depth is multiplied with these factors to estimate:

            - Minimum cell width (`fact_min`, default 0.2)
            - Domain-start (`fact_neg`, default 5), and
            - Domain-end (`fact_pos`, defaults to `fact_neg`).


    Returns
    -------

    h_min : float
        Minimum cell width.

    domain : list
        Start- and end-points of computation domain.

    """
    warnings.warn(MESHWARNING, FutureWarning)

    # Set fact_pos to fact_neg if not provided.
    if fact_pos is None:
        fact_pos = fact_neg

    # Get skin depth.
    skind = skin_depth(freq, 1/res)

    # Estimate minimum cell width.
    h_min = fact_min*skind
    if min_width is not None:  # Respect user input.
        if np.array(min_width).size == 1:
            h_min = min_width
        else:
            h_min = np.clip(h_min, *min_width)

    # Estimate computation domain.
    domain = [x0-fact_neg*skind, x0+fact_pos*skind]
    if limits is not None:  # Respect user input.
        domain = [min(limits[0], domain[0]), max(limits[1], domain[1])]

    return h_min, domain


def get_hx(alpha, domain, nx, x0, resp_domain=True):
    r"""Return cell widths for given input.

    Find the number of cells left (``nl``) and right (``nr``) of the center
    ``x0`` for the provided alpha. For this, we solve

    .. math::   \frac{x_\text{max}-x_0}{x_0-x_\text{min}} =
                \frac{a^\text{nr}-1}{a^\text{nl}-1}

    where :math:`a = 1+\alpha`.


    Parameters
    ----------

    alpha : float
        Stretching factor `a` is given by ``a=1+alpha``.

    domain : list
        ``[x_min, x_max]`` of model domain.

    nx : int
        Number of cells.

    x0 : float
        Center of the grid. ``x0`` is restricted to ``domain``.

    resp_domain : bool
        If False, then the domain-end might shift slightly to assure that the
        same stretching factor is applied throughout. If set to True (default),
        however, the domain is respected absolutely. This will introduce one
        stretch-factor which is different from the other stretch factors, to
        accommodate the restriction. This one-off factor is between the left-
        and right-side of `x0`, or, if `x1` is provided, just after `x1`.


    Returns
    -------
    hx : ndarray
        Cell widths of mesh. All points are given by
        ``np.r_[xmin, xmin+np.cumsum(hx)]``

    """
    warnings.warn(MESHWARNING, FutureWarning)

    if alpha <= 0.:  # If alpha <= 0: equal spacing (no stretching at all)
        hx = np.ones(nx)*np.diff(np.squeeze(domain))/nx

    else:            # Get stretched hx
        a = alpha+1

        # Get hx depending if x0 is on the domain boundary or not.
        if np.isclose(x0, domain[0]) or np.isclose(x0, domain[1]):
            # Get all a's
            alr = np.diff(domain)*alpha/(a**nx-1)*a**np.arange(nx)
            if x0 == domain[1]:
                alr = alr[::-1]

            # Compute differences
            hx = alr*np.diff(domain)/sum(alr)

        else:
            # Find number of elements left and right by solving:
            #     (xmax-x0)/(x0-xmin) = a**nr-1/(a**nl-1)
            nr = np.arange(2, nx+1)
            er = (domain[1]-x0)/(x0-domain[0]) - (a**nr[::-1]-1)/(a**nr-1)
            nl = np.argmin(abs(np.floor(er)))+1
            nr = nx-nl

            # Get all a's
            al = a**np.arange(nl-1, -1, -1)
            ar = a**np.arange(1, nr+1)

            # Compute differences
            if resp_domain:
                # This version honours domain[0] and domain[1], but to achieve
                # this it introduces one stretch-factor which is different from
                # all the others between al to ar.
                hx = np.r_[al*(x0-domain[0])/sum(al),
                           ar*(domain[1]-x0)/sum(ar)]
            else:
                # This version moves domain[1], but each stretch-factor is
                # exactly the same.
                fact = (x0-domain[0])/sum(al)  # Take distance from al.
                hx = np.r_[al, ar]*fact

                # Note: this hx is equivalent as providing the following h
                # to TensorMesh:
                # h = [(min_width, nl-1, -a), (min_width, n_nos+1),
                #      (min_width, nr, a)]

    return hx



"""
Everything to create model-properties for the multigrid solver.
"""
# Copyright 2018-2021 The emg3d Developers.
#
# This file is part of emg3d.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.  You may obtain a copy
# of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.

from copy import deepcopy

import numpy as np
from scipy.constants import epsilon_0

from emg3d import maps

__all__ = ['Model', 'VolumeModel']


# MODEL
class Model:
    r"""Create a model instance.

    Class to provide model parameters (x-, y-, and z-directed properties
    [resistivity or conductivity; linear or on log_10/log_e-scale], electric
    permittivity and magnetic permeability) to the solver. Relative magnetic
    permeability :math:`\mu_\mathrm{r}` is by default set to one and electric
    permittivity :math:`\varepsilon_\mathrm{r}` is by default set to zero, but
    they can also be provided (isotropically). Keep in mind that the multigrid
    method as implemented in `emg3d` only works for the diffusive
    approximation. As soon as the displacement-part in the Maxwell's equations
    becomes too dominant it will fail (high frequencies or very high electric
    permittivity).


    Parameters
    ----------
    grid : TensorMesh
        Grid on which to apply model.

    property_{x;y;z} : float or ndarray; default to 1.
        Material property in x-, y-, and z-directions. If ndarray, they must
        have the shape of grid.vnC (F-ordered) or grid.nC.

        By default, property refers to electrical resistivity. However, this
        can be changed with an appropriate map. For more info, see the
        description of the parameter `mapping`. The internals of `emg3d` work,
        irrelevant of the map, with electrical conductivities.

        Resistivities and conductivities have to be bigger than zero and
        smaller than infinity (if provided on a linear scale; not on
        logarithmic scales).

    mu_r : None, float, or ndarray
        Relative magnetic permeability (isotropic). If ndarray it must have the
        shape of grid.vnC (F-ordered) or grid.nC. Default is None, which
        corresponds to 1., but avoids the computation of zeta. Magnetic
        permeability has to be bigger than zero and smaller than infinity.

    epsilon_r : None, float, or ndarray
        Relative electric permittivity (isotropic). If ndarray it must have the
        shape of grid.vnC (F-ordered) or grid.nC. The displacement part is
        completely neglected (diffusive approximation) if set to None, which is
        the default. Electric permittivity has to be bigger than zero and
        smaller than infinity.

    mapping : str
        Defines what type the input `property_{x;y;z}`-values correspond to. By
        default, they represent resistivities (Ohm.m). The implemented types
        are:

        - 'Conductivity'; σ (S/m),
        - 'LgConductivity'; log_10(σ),
        - 'LnConductivity'; log_e(σ),
        - 'Resistivity'; ρ (Ohm.m); Default,
        - 'LgResistivity'; log_10(ρ),
        - 'LnResistivity'; log_e(ρ).

    """

    def __init__(self, grid, property_x=1., property_y=None, property_z=None,
                 mu_r=None, epsilon_r=None, mapping='Resistivity', **kwargs):
        """Initiate a new model."""

        # Ensure no kwargs left.
        if kwargs:
            raise TypeError(f"Unexpected **kwargs: {list(kwargs.keys())}")

        # Store required info from grid.
        if hasattr(grid, 'shape'):
            # This is an alternative possibility. Instead of the grid, we only
            # need the model.vnC. Mainly used internally to construct new
            # models.
            self.vnC = grid
            self.nC = np.prod(grid)
        else:
            self.nC = grid.nC
            self.vnC = grid.vnC

        # Copies of vnC and nC, but more widely used/known
        # (vnC and nC are the discretize attributes).
        self.shape = tuple(self.vnC)
        self.size = self.nC

        # Check case.
        self.case_names = ['isotropic', 'HTI', 'VTI', 'tri-axial']
        if property_y is None and property_z is None:
            # 0: Isotropic.
            self.case = 0
        elif property_z is None:
            # 1: HTI.
            self.case = 1
        elif property_y is None:
            # 2: VTI.
            self.case = 2
        else:
            # 3: Tri-axial anisotropy.
            self.case = 3

        # Get map.
        self.map = getattr(maps, 'Map'+mapping)()

        # Initiate all parameters.
        self._property_x = self._check_parameter(
                property_x, 'property_x', True)
        self._property_y = self._check_parameter(
                property_y, 'property_y', True)
        self._property_z = self._check_parameter(
                property_z, 'property_z', True)
        self._mu_r = self._check_parameter(mu_r, 'mu_r')
        self._epsilon_r = self._check_parameter(epsilon_r, 'epsilon_r')

    def __repr__(self):
        """Simple representation."""
        return (f"Model [{self.map.description}]; {self.case_names[self.case]}"
                f"{'' if self.mu_r is None else '; mu_r'}"
                f"{'' if self.epsilon_r is None else '; epsilon_r'}"
                f"; {self.vnC[0]} x {self.vnC[1]} x {self.vnC[2]} "
                f"({self.nC:,})")

    def __add__(self, model):
        """Add two models."""

        # Ensure model is a Model instance.
        if model.__class__.__name__ != 'Model':
            return NotImplemented

        # Check input.
        self._operator_test(model)

        # Apply operator.
        kwargs = self._apply_operator(model, np.add)

        # Return new Model instance.
        return Model(grid=np.array(self.vnC), **kwargs)

    def __sub__(self, model):
        """Subtract two models."""

        # Ensure model is a Model instance.
        if model.__class__.__name__ != 'Model':
            return NotImplemented

        # Check input.
        self._operator_test(model)

        # Apply operator.
        kwargs = self._apply_operator(model, np.subtract)

        # Return new Model instance.
        return Model(grid=np.array(self.vnC), **kwargs)

    def __eq__(self, model):
        """Compare two models.

        Note: Shape of parameters can be different, e.g. float, nC, or vnC. As
              long as all values agree it returns True.

        """

        # Check if model is a Model instance.
        equal = model.__class__.__name__ == 'Model'

        # Check input.
        if equal:
            try:
                _ = self._operator_test(model)
            except ValueError:
                equal = False

        # Compare values.
        if equal:
            equal *= np.allclose(self.property_x, model.property_x)
            equal *= np.allclose(self.property_y, model.property_y)
            equal *= np.allclose(self.property_z, model.property_z)
            # operator_test ensures mu_r are both or neither None.
            if self.mu_r is not None:
                equal *= np.allclose(self.mu_r, model.mu_r)
            # operator_test ensures epsilon_r are both or neither None.
            if self.epsilon_r is not None:
                equal *= np.allclose(self.epsilon_r, model.epsilon_r)

        return bool(equal)

    def copy(self):
        """Return a copy of the Model."""
        return self.from_dict(self.to_dict(True))

    def to_dict(self, copy=False):
        """Store the necessary information of the Model in a dict."""
        # Initiate dict.
        out = {}

        # Properties.
        out['property_x'] = self.property_x
        if self.case in [1, 3]:
            out['property_y'] = self.property_y
        else:
            out['property_y'] = None
        if self.case in [2, 3]:
            out['property_z'] = self.property_z
        else:
            out['property_z'] = None

        # mu_r.
        if self.mu_r is not None:
            out['mu_r'] = self.mu_r
        else:
            out['mu_r'] = None

        # epsilon_r.
        if self.epsilon_r is not None:
            out['epsilon_r'] = self.epsilon_r
        else:
            out['epsilon_r'] = None

        # vnC.
        out['vnC'] = np.array(self.vnC)

        # Map.
        out['mapping'] = self.map.name

        # Name
        out['__class__'] = self.__class__.__name__

        if copy:
            return deepcopy(out)
        else:
            return out

    @classmethod
    def from_dict(cls, inp):
        """Convert the dictionary into a Model instance.

        Parameters
        ----------
        inp : dict
            Dictionary as obtained from :func:`Model.to_dict`.
            The dictionary needs the keys `property_x`, `property_y`,
            `property_z`, `mu_r`, `epsilon_r`, `vnC`, and `mapping`.

        Returns
        -------
        obj : :class:`Model` instance

        """
        try:
            return cls(grid=inp['vnC'], property_x=inp['property_x'],
                       property_y=inp['property_y'],
                       property_z=inp['property_z'], mu_r=inp['mu_r'],
                       epsilon_r=inp['epsilon_r'], mapping=inp['mapping'])
        except KeyError as e:
            raise KeyError(f"Variable {e} missing in `inp`.") from e

    # PROPERTY
    @property
    def property_x(self):
        r"""Property in x-direction."""
        return self._return_parameter(self._property_x)

    @property_x.setter
    def property_x(self, property_x):
        r"""Update property in x-direction."""
        self._property_x = self._check_parameter(
                property_x, 'property_x', True)

    @property
    def property_y(self):
        r"""Property in y-direction."""
        if self.case in [1, 3]:  # HTI or tri-axial.
            return self._return_parameter(self._property_y)
        else:                    # Return property_x.
            return self._return_parameter(self._property_x)

    @property_y.setter
    def property_y(self, property_y):
        r"""Update property in y-direction."""

        # Adjust case in case property_z was not set so far.
        if self.case == 0:  # If it was isotropic, it is HTI now.
            self.case = 1
        elif self.case == 2:  # If it was VTI, it is tri-axial now.
            self.case = 3

        # Update it.
        self._property_y = self._check_parameter(
                property_y, 'property_y', True)

    @property
    def property_z(self):
        r"""Property in z-direction."""
        if self.case in [2, 3]:  # VTI or tri-axial.
            return self._return_parameter(self._property_z)
        else:                    # Return property_x.
            return self._return_parameter(self._property_x)

    @property_z.setter
    def property_z(self, property_z):
        r"""Update property in z-direction."""

        # Adjust case in case property_z was not set so far.
        if self.case == 0:  # If it was isotropic, it is VTI now.
            self.case = 2
        elif self.case == 1:  # If it was HTI, it is tri-axial now.
            self.case = 3

        # Update it.
        self._property_z = self._check_parameter(
                property_z, 'property_z', True)

    # MAGNETIC PERMEABILITIES
    @property
    def mu_r(self):
        r"""Magnetic permeability."""
        return self._return_parameter(self._mu_r)

    @mu_r.setter
    def mu_r(self, mu_r):
        r"""Update magnetic permeability."""
        self._mu_r = self._check_parameter(mu_r, 'mu_r')

    # ELECTRIC PERMITTIVITIES
    @property
    def epsilon_r(self):
        r"""Electric permittivity."""
        # Get epsilon.
        return self._return_parameter(self._epsilon_r)

    @epsilon_r.setter
    def epsilon_r(self, epsilon_r):
        r"""Update electric permittivity."""
        self._epsilon_r = self._check_parameter(epsilon_r, 'epsilon_r')

    def interpolate2grid(self, grid, new_grid, **grid2grid_opts):
        """Interpolate `Model` located on `grid` to `new_grid`.


        Parameters
        ----------
        grid, new_grid : TensorMesh
            Input and output model grids;
            :class:`emg3d.meshes.TensorMesh` instances.

        grid2grid_opts : dict
            Passed through to :func:`maps.grid2grid`. Defaults are
            `method='volume'`, `log=True`, and `extrapolate=True`.


        Returns
        -------
        NewModel : Model
            New :class:`Model` instance on `new_grid`.

        """

        # Get solver options, set to defaults if not provided.
        inp = {
                'method': 'volume',
                'extrapolate': True,
                'log': not self.map.name.startswith('L'),
                **(grid2grid_opts if grid2grid_opts is not None else {}),
                'grid': grid,
                'new_grid': new_grid
        }

        def ensure_vnc(prop):
            """Expand float-properties to shape vnC."""
            return prop*np.ones(grid.vnC) if prop.size == 1 else prop

        # property_x (always).
        property_x = maps.grid2grid(values=ensure_vnc(self.property_x), **inp)

        # property_y.
        if self.case in [1, 3]:
            property_y = maps.grid2grid(
                    values=ensure_vnc(self.property_y), **inp)
        else:
            property_y = None

        # property_z.
        if self.case in [2, 3]:
            property_z = maps.grid2grid(
                    values=ensure_vnc(self.property_z), **inp)
        else:
            property_z = None

        # mu_r.
        if self._mu_r is not None:
            mu_r = maps.grid2grid(values=ensure_vnc(self.mu_r), **inp)
        else:
            mu_r = None

        # epsilon_r.
        if self._epsilon_r is not None:
            epsilon_r = maps.grid2grid(
                values=ensure_vnc(self.epsilon_r), **inp)
        else:
            epsilon_r = None

        # Assemble coarse model.
        return Model(new_grid, property_x=property_x, property_y=property_y,
                     property_z=property_z, mu_r=mu_r, epsilon_r=epsilon_r,
                     mapping=self.map.name)

    # INTERNAL UTILITIES
    def _check_parameter(self, var, name, mapped=False):
        """Check parameter.

        - Shape must be (), (1,), nC, or vnC.
        - Value(s) must be 0 < var < inf.
        """

        # If None, exit.
        if var is None:
            return None

        # Cast it to floats, ravel.
        var = np.asarray(var, dtype=np.float64).ravel('F')

        # Check for wrong size.
        if var.size not in [1, self.nC]:
            raise ValueError(
                    f"Shape of {name} must be (), {self.vnC}, or "
                    f"{self.nC}.\nProvided: {var.shape}.")

        # Check they are positive.
        if not mapped or (mapped and not self.map.name.startswith('L')):
            if not np.all(var > 0):
                raise ValueError(f"`{name}` must be all `0 < var`.")

        # Check |val| < inf.
        if not np.all(np.abs(var) < np.inf):
            raise ValueError(f"`{name}` must be all `var < inf`.")

        return var

    def _return_parameter(self, var):
        """Return parameter as float or shape vnC."""

        # Return depending on value and size.
        if var is None:      # Because of mu_r, epsilon_r.
            return None
        elif var.size == 1:  # In case of float.
            return var
        else:                # Else, has shape vnC.
            return var.reshape(self.vnC, order='F')

    def _operator_test(self, model):
        """Check if `self` and `model` are consistent for operations.

        Note: {hx; hy; hz} is not checked. As long as the models have the
              same shape, isotropy, and mapping the operation will be carried
              out.

        """

        # Ensure the two instances have the same dimension.
        if np.any(self.shape != model.shape):
            msg = (f"Models could not be broadcast together with shapes "
                   f"{self.shape} and {model.shape}.")
            raise ValueError(msg)

        # Ensure the two instances have the same case.
        if self.case != model.case:
            msg = ("Models must be of the same isotropy type but have types"
                   f" '{self.case_names[self.case]}' and"
                   f" '{model.case_names[model.case]}'.")
            raise ValueError(msg)

        # Ensure both or none has mu_r:
        if hasattr(self.mu_r, 'dtype') != hasattr(model.mu_r, 'dtype'):
            msg = ("Either both or none of the models must have `mu_r` "
                   f"defined; provided: '{hasattr(self.mu_r, 'dtype')}' "
                   f"and '{hasattr(model.mu_r, 'dtype')}'.")
            raise ValueError(msg)

        # Ensure both or none has epsilon_r:
        if (hasattr(self.epsilon_r, 'dtype') !=
                hasattr(model.epsilon_r, 'dtype')):
            msg = ("Either both or none of the models must have `epsilon_r` "
                   f"defined; provided: '{hasattr(self.epsilon_r, 'dtype')}' "
                   f"and '{hasattr(model.epsilon_r, 'dtype')}'.")
            raise ValueError(msg)

        # Ensure the two instances have the same mapping:
        if self.map.name != model.map.name:
            msg = ("Models must have the same mapping but have mappings"
                   f" '{self.map.name}' and '{model.map.name}'.")
            raise ValueError(msg)

    def _apply_operator(self, model, operator):
        """Apply the provided operator to self and model."""

        kwargs = {}

        # Subtract properties.
        kwargs['property_x'] = operator(self.property_x, model.property_x)
        if self.case in [1, 3]:
            kwargs['property_y'] = operator(self.property_y, model.property_y)
        else:
            kwargs['property_y'] = None
        if self.case in [2, 3]:
            kwargs['property_z'] = operator(self.property_z, model.property_z)
        else:
            kwargs['property_z'] = None

        # Subtract mu_r.
        if self.mu_r is not None:
            kwargs['mu_r'] = operator(self.mu_r, model.mu_r)
        else:
            kwargs['mu_r'] = None

        # Subtract epsilon_r.
        if self.epsilon_r is not None:
            kwargs['epsilon_r'] = operator(self.epsilon_r, model.epsilon_r)
        else:
            kwargs['epsilon_r'] = None

        kwargs['mapping'] = self.map.name

        return kwargs


class VolumeModel:
    r"""Return a volume-averaged version of provided model.

    Takes a Model instance and returns the volume averaged values. This is used
    by the solver internally.

    .. math::

        \eta_{\{x,y,z\}} = -V\mathrm{i}\omega\mu_0
              \left(\rho^{-1}_{\{x,y,z\}} + \mathrm{i}\omega\varepsilon\right)

    .. math::

        \zeta = V\mu_\mathrm{r}^{-1}


    Parameters
    ----------
    grid : TensorMesh
        Grid on which to apply model.

    model : Model
        Model to transform to volume-averaged values.

    sfield : SourceField
       A VolumeModel is frequency-dependent. The frequency-information is taken
       from the provided source filed.

    """

    def __init__(self, grid, model, sfield):
        """Initiate a new model with volume-averaged properties."""

        # Store case, for restriction.
        self.case = model.case

        # eta_x
        self._eta_x = self.calculate_eta('property_x', grid, model, sfield)

        # eta_y
        if model.case in [1, 3]:  # HTI or tri-axial.
            self._eta_y = self.calculate_eta('property_y', grid, model, sfield)

        # eta_z
        if self.case in [2, 3]:  # VTI or tri-axial.
            self._eta_z = self.calculate_eta('property_z', grid, model, sfield)

        # zeta
        self._zeta = self.calculate_zeta('mu_r', grid, model)

    # ETA's
    @property
    def eta_x(self):
        r"""eta in x-direction."""
        return self._eta_x

    @property
    def eta_y(self):
        r"""eta in y-direction."""
        if self.case in [1, 3]:  # HTI or tri-axial.
            return self._eta_y
        else:                    # Return eta_x.
            return self._eta_x

    @property
    def eta_z(self):
        r"""eta in z-direction."""
        if self.case in [2, 3]:  # VTI or tri-axial.
            return self._eta_z
        else:                    # Return eta_x.
            return self._eta_x

    @property
    def zeta(self):
        r"""zeta."""
        return self._zeta

    @staticmethod
    def calculate_eta(name, grid, model, field):
        r"""eta: volume multiplied with conductivity."""

        # Initiate eta
        eta = field.smu0*grid.cell_volumes.reshape(grid.vnC, order='F')

        # Compute eta depending on epsilon.
        if model.epsilon_r is None:  # Diffusive approximation.
            eta *= model.map.backward(getattr(model, name))

        else:
            eps_term = field.sval*epsilon_0*model.epsilon_r
            sig_term = model.map.backward(getattr(model, name))
            eta *= sig_term - eps_term

        return eta

    @staticmethod
    def calculate_zeta(name, grid, model):
        r"""zeta: volume divided by mu_r."""

        zeta = grid.cell_volumes.reshape(grid.vnC, order='F')
        if getattr(model, name, None) is None:
            return zeta

        else:
            return zeta/getattr(model, name)



"""
Everything related to the multigrid solver that is a field: source field,
electric and magnetic fields, and fields at receivers.
"""
# Copyright 2018-2021 The emg3d Developers.
#
# This file is part of emg3d.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.  You may obtain a copy
# of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.

import warnings
from copy import deepcopy

import numpy as np
from scipy.constants import mu_0
from scipy.special import sindg, cosdg

from emg3d import maps, models, utils

__all__ = ['Field', 'SourceField', 'get_source_field', 'get_receiver',
           'get_receiver_response', 'get_h_field']


class Field(np.ndarray):
    r"""Create a Field instance with x-, y-, and z-views of the field.

    A `Field` is an `ndarray` with additional views of the x-, y-, and
    z-directed fields as attributes, stored as `fx`, `fy`, and `fz`. The
    default array contains the whole field, which can be the electric field,
    the source field, or the residual field, in a 1D array. A `Field` instance
    has additionally the property `ensure_pec` which, if called, ensures
    Perfect Electric Conductor (PEC) boundary condition. It also has the two
    attributes `amp` and `pha` for the amplitude and phase, as common in
    frequency-domain CSEM.

    A `Field` can be initiated in three ways:

    1. ``Field(grid, dtype=np.complex128)``:
       Calling it with a :class:`emg3d.meshes.TensorMesh` instance returns a
       `Field` instance of correct dimensions initiated with zeroes of data
       type `dtype`.
    2. ``Field(grid, field)``:
       Calling it with a :class:`emg3d.meshes.TensorMesh` instance and an
       `ndarray` returns a `Field` instance of the provided `ndarray`, of same
       data type.
    3. ``Field(fx, fy, fz)``:
       Calling it with three `ndarray`'s which represent the field in x-, y-,
       and z-direction returns a `Field` instance with these views, of same
       data type.

    Sort-order is 'F'.


    Parameters
    ----------

    fx_or_grid : :class:`emg3d.meshes.TensorMesh` or ndarray
        Either a TensorMesh instance or an ndarray of shape grid.nEx or
        grid.vnEx. See explanations above. Only mandatory parameter; if the
        only one provided, it will initiate a zero-field of `dtype`.

    fy_or_field : :class:`Field` or ndarray, optional
        Either a Field instance or an ndarray of shape grid.nEy or grid.vnEy.
        See explanations above.

    fz : ndarray, optional
        An ndarray of shape grid.nEz or grid.vnEz. See explanations above.

    dtype : dtype, optional
        Only used if ``fy_or_field=None`` and ``fz=None``; the initiated
        zero-field for the provided TensorMesh has data type `dtype`.
        Default: complex.

    freq : float, optional
        Source frequency (Hz), used to compute the Laplace parameter `s`.
        Either positive or negative:

        - `freq` > 0: Frequency domain, hence
          :math:`s = -\mathrm{i}\omega = -2\mathrm{i}\pi f` (complex);
        - `freq` < 0: Laplace domain, hence
          :math:`s = f` (real).

        Just added as info if provided.

    """

    def __new__(cls, fx_or_grid, fy_or_field=None, fz=None,
                dtype=np.complex128, freq=None):
        """Initiate a new Field instance."""

        # Collect field
        if fy_or_field is None and fz is None:          # Empty Field with
            new = np.zeros(fx_or_grid.nE, dtype=dtype)  # dimension grid.nE.
        elif fz is None:                  # grid and field provided
            new = fy_or_field
        else:                             # fx, fy, fz provided
            new = np.r_[fx_or_grid.ravel('F'), fy_or_field.ravel('F'),
                        fz.ravel('F')]

        # Store the field as object
        obj = np.asarray(new).view(cls)

        # Store relevant numbers for the views.
        if fy_or_field is not None and fz is not None:  # Deduce from arrays
            obj.nEx = fx_or_grid.size
            obj.nEy = fy_or_field.size
            obj.nEz = fz.size
            obj.vnEx = fx_or_grid.shape
            obj.vnEy = fy_or_field.shape
            obj.vnEz = fz.shape
        else:                                     # If grid is provided
            attr_list = ['nEx', 'nEy', 'nEz', 'vnEx', 'vnEy', 'vnEz']

            for attr in attr_list:
                setattr(obj, attr, getattr(fx_or_grid, attr))

            # Ensure the grid has three dimensions.
            # (Can happend with a 1D or 2D discretize mesh.)
            if None in [obj.nEx, obj.nEy, obj.nEz]:
                raise ValueError("Provided grid must be a 3D grid.")

        # Store frequency
        if freq is None and hasattr(fy_or_field, 'freq'):
            freq = fy_or_field._freq
        obj._freq = freq
        if freq == 0.0:
            raise ValueError(
                    "`freq` must be >0 (frequency domain) "
                    "or <0 (Laplace domain).\n"
                    f"Provided frequency: {freq} Hz.")

        return obj

    def __array_finalize__(self, obj):
        """Ensure relevant numbers are stored no matter how created."""
        if obj is None:
            return

        self.nEx = getattr(obj, 'nEx', None)
        self.nEy = getattr(obj, 'nEy', None)
        self.nEz = getattr(obj, 'nEz', None)
        self.vnEx = getattr(obj, 'vnEx', None)
        self.vnEy = getattr(obj, 'vnEy', None)
        self.vnEz = getattr(obj, 'vnEz', None)
        self._freq = getattr(obj, '_freq', None)

    def __reduce__(self):
        """Customize __reduce__ to make `Field` work with pickle.
        => https://stackoverflow.com/a/26599346
        """
        # Get the parent's __reduce__ tuple.
        pickled_state = super().__reduce__()

        # Create our own tuple to pass to __setstate__.
        new_state = pickled_state[2]
        attr_list = ['nEx', 'nEy', 'nEz', 'vnEx', 'vnEy', 'vnEz', '_freq']
        for attr in attr_list:
            new_state += (getattr(self, attr),)

        # Return tuple that replaces parent's __setstate__ tuple with our own.
        return (pickled_state[0], pickled_state[1], new_state)

    def __setstate__(self, state):
        """Customize __setstate__ to make `Field` work with pickle.
        => https://stackoverflow.com/a/26599346
        """
        # Set the necessary attributes (in reverse order).
        attr_list = ['nEx', 'nEy', 'nEz', 'vnEx', 'vnEy', 'vnEz', '_freq']
        attr_list.reverse()
        for i, name in enumerate(attr_list):
            i += 1  # We need it 1..#attr instead of 0..#attr-1.
            setattr(self, name, state[-i])

        # Call the parent's __setstate__ with the other tuple elements.
        super().__setstate__(state[0:-i])

    def copy(self):
        """Return a copy of the Field."""
        return self.from_dict(self.to_dict(True))

    def to_dict(self, copy=False):
        """Store the necessary information of the Field in a dict."""
        out = {'field': np.array(self.field), 'freq': self._freq,
               'vnEx': self.vnEx, 'vnEy': self.vnEy, 'vnEz': self.vnEz,
               '__class__': self.__class__.__name__}
        if copy:
            return deepcopy(out)
        else:
            return out

    @classmethod
    def from_dict(cls, inp):
        """Convert dictionary into :class:`Field` instance.

        Parameters
        ----------
        inp : dict
            Dictionary as obtained from :func:`Field.to_dict`.
            The dictionary needs the keys `field`, `freq`, `vnEx`, `vnEy`, and
            `vnEz`.

        Returns
        -------
        obj : :class:`Field` instance

        """

        # Create a dummy with the required attributes for the field instance.
        class Grid:
            pass

        grid = Grid()

        # Check and get the required keys from the input.
        try:
            field = inp['field']
            freq = inp['freq']
            grid.vnEx = inp['vnEx']
            grid.vnEy = inp['vnEy']
            grid.vnEz = inp['vnEz']
        except KeyError as e:
            raise KeyError(f"Variable {e} missing in `inp`.") from e

        # Compute missing info.
        grid.nEx = np.prod(grid.vnEx)
        grid.nEy = np.prod(grid.vnEy)
        grid.nEz = np.prod(grid.vnEz)
        grid.nE = grid.nEx + grid.nEy + grid.nEz

        # Return Field instance.
        return cls(fx_or_grid=grid, fy_or_field=field, freq=freq)

    @property
    def field(self):
        """Entire field, 1D [fx, fy, fz]."""
        return self.view()

    @field.setter
    def field(self, field):
        """Update field, 1D [fx, fy, fz]."""
        self.view()[:] = field

    @property
    def fx(self):
        """View of the x-directed field in the x-direction (nCx, nNy, nNz)."""
        return self.view()[:self.nEx].reshape(self.vnEx, order='F')

    @fx.setter
    def fx(self, fx):
        """Update field in x-direction."""
        self.view()[:self.nEx] = fx.ravel('F')

    @property
    def fy(self):
        """View of the field in the y-direction (nNx, nCy, nNz)."""
        return self.view()[self.nEx:-self.nEz].reshape(self.vnEy, order='F')

    @fy.setter
    def fy(self, fy):
        """Update field in y-direction."""
        self.view()[self.nEx:-self.nEz] = fy.ravel('F')

    @property
    def fz(self):
        """View of the field in the z-direction (nNx, nNy, nCz)."""
        return self.view()[-self.nEz:].reshape(self.vnEz, order='F')

    @fz.setter
    def fz(self, fz):
        """Update electric field in z-direction."""
        self.view()[-self.nEz:] = fz.ravel('F')

    def amp(self):
        """Amplitude of the electromagnetic field."""
        return utils.EMArray(self.view()).amp()

    def pha(self, deg=False, unwrap=True, lag=True):
        """Phase of the electromagnetic field.

        Parameters
        ----------
        deg : bool
            If True the returned phase is in degrees, else in radians.
            Default is False (radians).

        unwrap : bool
            If True the returned phase is unwrapped.
            Default is True (unwrapped).

        lag : bool
            If True the returned phase is lag, else lead defined.
            Default is True (lag defined).

        """
        return utils.EMArray(self.view()).pha(deg, unwrap, lag)

    @property
    def freq(self):
        """Return frequency."""
        if self._freq is None:
            return None
        else:
            return abs(self._freq)

    @property
    def smu0(self):
        """Return s*mu_0; mu_0 = Magn. permeability of free space [H/m]."""
        if getattr(self, '_smu0', None) is None:
            if self.sval is not None:
                self._smu0 = self.sval*mu_0
            else:
                self._smu0 = None

        return self._smu0

    @property
    def sval(self):
        """Return s; s=iw in frequency domain; s=freq in Laplace domain."""

        if getattr(self, '_sval', None) is None:
            if self._freq is not None:
                if self._freq < 0:  # Laplace domain; s.
                    self._sval = np.array(self._freq)
                else:  # Frequency domain; s = iw = 2i*pi*f.
                    self._sval = np.array(-2j*np.pi*self._freq)
            else:
                self._sval = None

        return self._sval

    @property
    def ensure_pec(self):
        """Set Perfect Electric Conductor (PEC) boundary condition."""
        # Apply PEC to fx
        self.fx[:, 0, :] = 0.
        self.fx[:, -1, :] = 0.
        self.fx[:, :, 0] = 0.
        self.fx[:, :, -1] = 0.

        # Apply PEC to fy
        self.fy[0, :, :] = 0.
        self.fy[-1, :, :] = 0.
        self.fy[:, :, 0] = 0.
        self.fy[:, :, -1] = 0.

        # Apply PEC to fz
        self.fz[0, :, :] = 0.
        self.fz[-1, :, :] = 0.
        self.fz[:, 0, :] = 0.
        self.fz[:, -1, :] = 0.

    @property
    def is_electric(self):
        """Returns True if Field is electric, False if it is magnetic."""
        return self.vnEx[0] < self.vnEy[0]


class SourceField(Field):
    r"""Create a Source-Field instance with x-, y-, and z-views of the field.

    A subclass of :class:`Field`. Additional properties are the real-valued
    source vector (`vector`, `vx`, `vy`, `vz`), which sum is always one. For a
    `SourceField` frequency is a mandatory  parameter, unlike for a `Field`
    (recommended also for `Field` though),

    Parameters
    ----------

    fx_or_grid : :class:`emg3d.meshes.TensorMesh` or ndarray
        Either a TensorMesh instance or an ndarray of shape grid.nEx or
        grid.vnEx. See explanations above. Only mandatory parameter; if the
        only one provided, it will initiate a zero-field of `dtype`.

    fy_or_field : :class:`Field` or ndarray, optional
        Either a Field instance or an ndarray of shape grid.nEy or grid.vnEy.
        See explanations above.

    fz : ndarray, optional
        An ndarray of shape grid.nEz or grid.vnEz. See explanations above.

    dtype : dtype, optional
        Only used if ``fy_or_field=None`` and ``fz=None``; the initiated
        zero-field for the provided TensorMesh has data type `dtype`.
        Default: complex.

    freq : float
        Source frequency (Hz), used to compute the Laplace parameter `s`.
        Either positive or negative:

        - `freq` > 0: Frequency domain, hence
          :math:`s = -\mathrm{i}\omega = -2\mathrm{i}\pi f` (complex);
        - `freq` < 0: Laplace domain, hence
          :math:`s = f` (real).

        In difference to `Field`, the frequency has to be provided for
        a `SourceField`.

    """

    def __new__(cls, fx_or_grid, fy_or_field=None, fz=None,
                dtype=np.complex128, freq=None):
        """Initiate a new Source Field."""
        # Ensure frequency is provided.
        if freq is None:
            raise ValueError("SourceField requires the frequency.")

        if freq > 0:
            dtype = complex
        else:
            dtype = float

        return super().__new__(cls, fx_or_grid, fy_or_field=fy_or_field,
                               fz=fz, dtype=dtype, freq=freq)

    @property
    def vector(self):
        """Entire vector, 1D [vx, vy, vz]."""
        return np.real(self.field/self.smu0)

    @property
    def vx(self):
        """View of the x-directed vector in the x-direction (nCx, nNy, nNz)."""
        return np.real(self.field.fx/self.smu0)

    @property
    def vy(self):
        """View of the vector in the y-direction (nNx, nCy, nNz)."""
        return np.real(self.field.fy/self.smu0)

    @property
    def vz(self):
        """View of the vector in the z-direction (nNx, nNy, nCz)."""
        return np.real(self.field.fz/self.smu0)


def get_source_field(grid, src, freq, strength=0, electric=True, length=1.0,
                     decimals=6, **kwargs):
    r"""Return the source field.

    The source field is given in Equation 2 in [Muld06]_,

    .. math::

        s \mu_0 \mathbf{J}_\mathrm{s} ,

    where :math:`s = \mathrm{i} \omega`. Either finite length dipoles,
    infinitesimal small point dipoles, or arbitrarily shaped segments can be
    defined, whereas the returned source field corresponds to a normalized
    (1 Am) source distributed within the cell(s) it resides (can be changed
    with the `strength`-parameter).

    The adjoint of the trilinear interpolation is used to distribute the
    point(s) to the grid edges, which corresponds to the discretization of a
    Dirac ([PlDM07]_).


    Parameters
    ----------
    grid : TensorMesh
        Model grid; a :class:`emg3d.meshes.TensorMesh` instance.

    src : list of floats
        Source coordinates (m). There are three formats:

          - Finite length dipole: ``[x0, x1, y0, y1, z0, z1]``.
          - Point dipole: ``[x, y, z, azimuth, dip]``.
          - Arbitrarily shaped source: ``[[x-coo], [y-coo], [z-coo]]``.

        Point dipoles will be converted internally to finite length dipoles
        of ``length``. In the case of a point dipole one can set
        ``electric=False``, which will create a square loop of
        ``length``x``length`` perpendicular to the dipole.

    freq : float
        Source frequency (Hz), used to compute the Laplace parameter `s`.
        Either positive or negative:

        - `freq` > 0: Frequency domain, hence
          :math:`s = -\mathrm{i}\omega = -2\mathrm{i}\pi f` (complex);
        - `freq` < 0: Laplace domain, hence
          :math:`s = f` (real).

    strength : float or complex, optional
        Source strength (A):

          - If 0, output is normalized to a source of 1 m length, and source
            strength of 1 A.
          - If != 0, output is returned for given source length and strength.

        Default is 0.

    electric : bool, optional
        Shortcut to create a magnetic source. If False, the format of ``src``
        must be that of a point dipole: ``[x, y, z, azimuth, dip]`` (for the
        other formats setting ``electric`` has no effect). It then creates a
        square loop perpendicular to this dipole, with side-length 1.
        Default is True, meaning an electric source.

    length : float, optional
        Length (m) of the point dipole when converted to a finite length
        dipole, or edge-length (m) of the square loop if ``electric=False``.
        Default is 1.0.

    decimals: int, optional
        Grid nodes and source coordinates are rounded to given number of
        decimals. Default is 6 (micrometer).


    Returns
    -------
    sfield : :func:`SourceField` instance
        Source field, normalized to 1 A m.

    """

    if 'msrc' in kwargs:
        msg = ("``msrc`` is deprecated and overwrites ``electric``. "
               "Use ``electric`` instead.\nAlso, the sign of the magnetic "
               "source was corrected (switched)")
        warnings.warn(msg, FutureWarning)
        electric = not kwargs.pop('msrc')

    # Ensure no kwargs left.
    if kwargs:
        raise TypeError(f"Unexpected **kwargs: {list(kwargs.keys())}")

    # Cast some parameters.
    if not np.allclose(np.size(src[0]), [np.size(c) for c in src]):
        raise ValueError(
                "All source coordinates must have the same dimension."
                f"Provided source: {src}.")

    src = np.asarray(src, dtype=np.float64)
    strength = np.asarray(strength)

    # Convert point dipole sources to finite dipoles or loops (electric).
    if src.shape == (5, ):  # Point dipole

        if not electric:  # Magnetic: convert to square loop perp. to dipole.
            src = _square_loop_from_point_dipole(src, length)
            # src.shape = (3, 5)

        else:  # Electric: convert to finite length.
            src = _finite_dipole_from_point_dipole(src, length)
            # src.shape = (6, )

    # Get arbitrary shaped sources recursively.
    if src.shape[0] == 3 and src.ndim > 1:

        # Get arbitrarily shaped dipole source using recursion.
        sx, sy, sz = src

        # Get normalized segment lengths.
        lengths = np.sqrt(np.sum((src[:, :-1] - src[:, 1:])**2, axis=0))
        if strength == 0:
            lengths /= lengths.sum()
        else:  # (Not in-place multiplication, as strength can be complex.)
            lengths = lengths*strength

        # Initiate a zero-valued source field and loop over segments.
        sfield = SourceField(grid, freq=freq)
        sfield.src = src
        sfield.strength = strength
        sfield.moment = np.array([0., 0, 0], dtype=lengths.dtype)

        # Loop over elements.
        for i in range(sx.size-1):
            segment = (sx[i], sx[i+1], sy[i], sy[i+1], sz[i], sz[i+1])
            seg_field = get_source_field(grid, segment, freq, lengths[i])
            sfield += seg_field
            sfield.moment += seg_field.moment

        # Check this with iw/-iw; source definition etc.
        if not electric:
            sfield *= -1

        return sfield

    # From here onwards `src` has to be a finite length dipole  of format
    # [x1, x2, y1, y2, z1, z2]. Ensure that:
    if src.shape != (6, ):
        raise ValueError(
                "Source is wrong defined. It must be either\n- a point, "
                "[x, y, z, azimuth, dip],\n- a finite dipole, "
                "[x1, x2, y1, y2, z1, z2], or\n- an arbitrarily shaped "
                "dipole, [[x-coo], [y-coo], [z-coo]].\n"
                f"Provided source: {src}.")

    # Get length in each direction.
    length = src[1::2]-src[::2]

    # Ensure finite length dipole is not a point dipole.
    if np.allclose(length, 0, atol=1e-15):
        raise ValueError("Provided finite dipole has no length; use "
                         "the format [x, y, z, azimuth, dip] instead.")

    # Get source moment (individually for x, y, z).
    if strength == 0:  # 1 A m
        length /= np.linalg.norm(length)
        moment = length
    else:              # Multiply source length with source strength
        moment = strength*length

    # Initiate zero source field.
    sfield = SourceField(grid, freq=freq)

    # Return source-field for each direction.
    for xyz, sf in enumerate([sfield.fx, sfield.fy, sfield.fz]):

        # Get source field for this direction.
        _finite_source_xyz(grid, src, sf, xyz, decimals)

        # Multiply by moment*s*mu
        sf *= moment[xyz]*sfield.smu0

    # Add src and moment information.
    sfield.src = src
    sfield.strength = strength
    sfield.moment = moment

    return sfield


def get_receiver(grid, values, coordinates, method='cubic', extrapolate=False):
    """Return values corresponding to grid at coordinates.

    Works for electric fields as well as magnetic fields obtained with
    :func:`get_h_field`, and for model parameters.

    Note that in order to avoid boundary effects the first and last value in
    each direction is neglected. Values for coordinates outside of the grid
    are set to NaN's, unless extrapolate is set.


    Parameters
    ----------
    grid : :class:`emg3d.meshes.TensorMesh`
        The model grid.

    values : ndarray
        Field instance, or a particular field (e.g. field.fx); Model
        parameters.

    coordinates : tuple (x, y, z)
        Coordinates (x, y, z) where to interpolate `values`; e.g. receiver
        locations.

    method : str, optional
        The method of interpolation to perform, 'linear' or 'cubic'.
        Default is 'cubic' (forced to 'linear' if there are less than 3 points
        in any direction).

    extrapolate : bool
        If True, points on `new_grid` which are outside of `grid` are
        filled by the nearest value (if ``method='cubic'``) or by extrapolation
        (if ``method='linear'``). If False, points outside are set to NaN's.

        Default is False.


    Returns
    -------
    new_values : ndarray or :class:`utils.EMArray`
        Values at `coordinates`.

        If input was a field it returns an EMArray, which is a subclassed
        ndarray with ``.pha`` and ``.amp`` attributes.

        If input was an entire Field instance, output is a tuple (fx, fy, fz).


    See Also
    --------
    grid2grid : Interpolation of model parameters or fields to a new grid.
    get_receiver_response : Get response for arbitrarily rotated receivers.

    """
    # If values is a Field instance, call it recursively for each field.
    if hasattr(values, 'field') and values.field.ndim == 1:
        fx = get_receiver(grid, values.fx, coordinates, method, extrapolate)
        fy = get_receiver(grid, values.fy, coordinates, method, extrapolate)
        fz = get_receiver(grid, values.fz, coordinates, method, extrapolate)
        return fx, fy, fz

    if len(coordinates) != 3:
        raise ValueError(
                "Coordinates needs to be in the form (x, y, z).\n"
                f"Length of provided coord.: {len(coordinates)}.")

    # Get the vectors corresponding to input data. Dimensions:
    #
    #         E-field          H-field      |  Model Parameter
    #  x: [nCx, nNy, nNz]  [nNx, nCy, nCz]  |
    #  y: [nNx, nCy, nNz]  [nCx, nNy, nCz]  |  [nCx, nCy, nCz]
    #  z: [nNx, nNy, nCz]  [nCx, nCy, nNz]  |
    #
    points = tuple()
    for i, coord in enumerate(['x', 'y', 'z']):
        if values.shape[i] == getattr(grid, 'shape_nodes')[i]:
            pts = (getattr(grid, 'nodes_'+coord)[1:-1], )
        else:
            pts = (getattr(grid, 'cell_centers_'+coord)[1:-1], )

        # Add to points.
        points += pts

    if extrapolate:
        fill_value = None
        mode = 'nearest'
    else:
        fill_value = np.array(0, values.dtype)*np.nan
        mode = 'constant'
    out = maps.interp3d(points, values[1:-1, 1:-1, 1:-1], coordinates,
                        method, fill_value, mode, cval=np.nan)

    # Return an EMArray if input is a field, else simply the values.
    if values.size == grid.nC:
        return out
    else:
        return utils.EMArray(out)


def get_receiver_response(grid, field, rec):
    """Return the field (response) at receiver coordinates.

    Note that in order to avoid boundary effects the first and last value in
    each direction is neglected. Field values for coordinates outside of the
    grid are set to NaN's.


    Parameters
    ----------
    grid : :class:`emg3d.meshes.TensorMesh`
        The model grid.

    field : :class:`Field`
        The electric or magnetic field.

    rec : tuple (x, y, z, azimuth, dip)
        Receiver coordinates and angles (m, °).

        All values can either be a scalar or having the same length as number
        of receivers.

        Angles:

        - azimuth (°): horizontal deviation from x-axis, anti-clockwise.
        - dip (°): vertical deviation from xy-plane up-wards.


    Returns
    -------
    responses : :class:`utils.EMArray`
        Responses at receiver.


    .. note::

        Currently only implemented for point receivers, not for finite length
        dipoles.


    See Also
    --------
    get_receiver : Get values at coordinates (fields and models).

    """

    # Check receiver dimension.
    if len(rec) != 5:
        raise ValueError(
                "`rec` needs to be in the form (x, y, z, azimuth, dip).\n"
                f"Length of provided `rec`: {len(rec)}.")

    # Check field dimension to ensure it is not a particular field.
    if field.field.ndim == 3:
        raise ValueError("`field` must be a `Field`-instance, not a\n"
                         "particular field such as `field.fx`.")

    # Get the vectors corresponding to input data.
    if field.is_electric:
        points = ((grid.cell_centers_x, grid.nodes_y, grid.nodes_z),
                  (grid.nodes_x, grid.cell_centers_y, grid.nodes_z),
                  (grid.nodes_x, grid.nodes_y, grid.cell_centers_z))
    else:
        points = ((grid.nodes_x, grid.cell_centers_y, grid.cell_centers_z),
                  (grid.cell_centers_x, grid.nodes_y, grid.cell_centers_z),
                  (grid.cell_centers_x, grid.cell_centers_y, grid.nodes_z))

    # Remove first and last value in each direction.
    points = tuple([tuple([p[1:-1] for p in pp]) for pp in points])

    # Pre-allocate the response.
    resp = np.zeros(max([np.atleast_1d(x).size for x in rec]),
                    dtype=field.dtype)

    # Add the required responses.
    factors = _rotation(*rec[3:])  # Geometrical weights from angles.
    inp = {'method': 'cubic', 'fill_value': 0.0, 'mode': 'constant',
           'cval': np.nan}
    for i, ff in enumerate((field.fx, field.fy, field.fz)):
        if np.any(abs(factors[i]) > 1e-10):
            resp += factors[i]*maps.interp3d(
                       points[i], ff[1:-1, 1:-1, 1:-1], rec[:3], **inp)

    # Return response.
    return utils.EMArray(resp)


def get_h_field(grid, model, field):
    r"""Return magnetic field corresponding to provided electric field.

    Retrieve the magnetic field :math:`\mathbf{H}` from the electric field
    :math:`\mathbf{E}` using Farady's law, given by

    .. math::

        \nabla \times \mathbf{E} = \rm{i}\omega\mu\mathbf{H} .

    Note that the magnetic field in x-direction is defined in the center of the
    face defined by the electric field in y- and z-directions, and similar for
    the other field directions. This means that the provided electric field and
    the returned magnetic field have different dimensions::

       E-field:  x: [grid.cell_centers_x,  grid.nodes_y,  grid.nodes_z]
                 y: [ grid.nodes_x, grid.cell_centers_y,  grid.nodes_z]
                 z: [ grid.nodes_x,  grid.nodes_y, grid.cell_centers_z]

       H-field:  x: [ grid.nodes_x, grid.cell_centers_y, grid.cell_centers_z]
                 y: [grid.cell_centers_x,  grid.nodes_y, grid.cell_centers_z]
                 z: [grid.cell_centers_x, grid.cell_centers_y,  grid.nodes_z]


    Parameters
    ----------
    grid : TensorMesh
        Model grid; :class:`TensorMesh` instance.

    model : Model
        Model; :class:`Model` instance.

    field : Field
        Electric field; :class:`Field` instance.


    Returns
    -------
    hfield : Field
        Magnetic field; :class:`Field` instance.

    """

    # Carry out the curl (^ corresponds to differentiation axis):
    # H_x = (E_z^1 - E_y^2)
    e3d_hx = (np.diff(field.fz, axis=1)/grid.h[1][None, :, None] -
              np.diff(field.fy, axis=2)/grid.h[2][None, None, :])

    # H_y = (E_x^2 - E_z^0)
    e3d_hy = (np.diff(field.fx, axis=2)/grid.h[2][None, None, :] -
              np.diff(field.fz, axis=0)/grid.h[0][:, None, None])

    # H_z = (E_y^0 - E_x^1)
    e3d_hz = (np.diff(field.fy, axis=0)/grid.h[0][:, None, None] -
              np.diff(field.fx, axis=1)/grid.h[1][None, :, None])

    # If relative magnetic permeability is not one, we have to take the volume
    # into account, as mu_r is volume-averaged.
    if model._mu_r is not None:

        # Get volume-averaged values.
        vmodel = models.VolumeModel(grid, model, field)

        # Plus and minus indices.
        ixm = np.r_[0, np.arange(grid.vnC[0])]
        ixp = np.r_[np.arange(grid.vnC[0]), grid.vnC[0]-1]
        iym = np.r_[0, np.arange(grid.vnC[1])]
        iyp = np.r_[np.arange(grid.vnC[1]), grid.vnC[1]-1]
        izm = np.r_[0, np.arange(grid.vnC[2])]
        izp = np.r_[np.arange(grid.vnC[2]), grid.vnC[2]-1]

        # Average mu_r for dual-grid.
        zeta_x = (vmodel.zeta[ixm, :, :] + vmodel.zeta[ixp, :, :])/2.
        zeta_y = (vmodel.zeta[:, iym, :] + vmodel.zeta[:, iyp, :])/2.
        zeta_z = (vmodel.zeta[:, :, izm] + vmodel.zeta[:, :, izp])/2.

        hvx = grid.h[0][:, None, None]
        hvy = grid.h[1][None, :, None]
        hvz = grid.h[2][None, None, :]

        # Define the widths of the dual grid.
        dx = (np.r_[0., grid.h[0]] + np.r_[grid.h[0], 0.])/2.
        dy = (np.r_[0., grid.h[1]] + np.r_[grid.h[1], 0.])/2.
        dz = (np.r_[0., grid.h[2]] + np.r_[grid.h[2], 0.])/2.

        # Multiply fields by mu_r.
        e3d_hx *= zeta_x/(dx[:, None, None]*hvy*hvz)
        e3d_hy *= zeta_y/(hvx*dy[None, :, None]*hvz)
        e3d_hz *= zeta_z/(hvx*hvy*dz[None, None, :])

    # Create a Field instance and divide by s*mu_0 and return.
    return -Field(e3d_hx, e3d_hy, e3d_hz)/field.smu0


def _finite_source_xyz(grid, src, s, xyz, decimals):
    """Set finite dipole source using the adjoint interpolation method.

    See :func:`get_source_field` for further details.

    """
    # Round nodes and src coordinates (to avoid floating point issues etc).
    nodes_x = np.round(grid.nodes_x, decimals)
    nodes_y = np.round(grid.nodes_y, decimals)
    nodes_z = np.round(grid.nodes_z, decimals)
    src = np.round(src, decimals)

    # Ensure source is within nodes.
    outside = (src[0] < nodes_x[0] or src[1] > nodes_x[-1] or
               src[2] < nodes_y[0] or src[3] > nodes_y[-1] or
               src[4] < nodes_z[0] or src[5] > nodes_z[-1])
    if outside:
        raise ValueError(f"Provided source outside grid: {src}.")

    # Source lengths in x-, y-, and z-directions.
    d_xyz = src[1::2]-src[::2]

    # Inverse source lengths.
    id_xyz = d_xyz.copy()
    id_xyz[id_xyz != 0] = 1/id_xyz[id_xyz != 0]

    # Cell fractions.
    a1 = (nodes_x-src[0])*id_xyz[0]
    a2 = (nodes_y-src[2])*id_xyz[1]
    a3 = (nodes_z-src[4])*id_xyz[2]

    # Get range of indices of cells in which source resides.
    def min_max_ind(vector, i):
        """Return [min, max]-index of cells in which source resides."""
        vmin = min(src[2*i:2*i+2])
        vmax = max(src[2*i:2*i+2])
        return [max(0, np.where(vmin < np.r_[vector, np.infty])[0][0]-1),
                max(0, np.where(vmax < np.r_[vector, np.infty])[0][0]-1)]

    rix = min_max_ind(nodes_x, 0)
    riy = min_max_ind(nodes_y, 1)
    riz = min_max_ind(nodes_z, 2)

    # Loop over these indices.
    for iz in range(riz[0], min(riz[1]+1, a3.size-1)):
        for iy in range(riy[0], min(riy[1]+1, a2.size-1)):
            for ix in range(rix[0], min(rix[1]+1, a1.size-1)):

                # Determine centre of gravity of line segment in cell.
                aa = np.vstack([[a1[ix], a1[ix+1]], [a2[iy], a2[iy+1]],
                                [a3[iz], a3[iz+1]]])
                aa = np.sort(aa[d_xyz != 0, :], 1)
                al = max(0, aa[:, 0].max())  # Left and right
                ar = min(1, aa[:, 1].min())  # elements.

                # Characteristics of this cell.
                xmin = src[::2]+al*d_xyz
                xmax = src[::2]+ar*d_xyz
                x_c = (xmin+xmax)/2.0
                slen = np.linalg.norm(src[1::2]-src[::2])
                x_len = np.linalg.norm(xmax-xmin)/slen

                # Contribution to edge (coordinate xyz)
                rx = (x_c[0]-nodes_x[ix])/grid.h[0][ix]
                ex = 1-rx
                ry = (x_c[1]-nodes_y[iy])/grid.h[1][iy]
                ey = 1-ry
                rz = (x_c[2]-nodes_z[iz])/grid.h[2][iz]
                ez = 1-rz

                # Add to field (only if segment inside cell).
                if min(rx, ry, rz) >= 0 and np.max(np.abs(ar-al)) > 0:

                    if xyz == 0:
                        s[ix, iy, iz] += ey*ez*x_len
                        s[ix, iy+1, iz] += ry*ez*x_len
                        s[ix, iy, iz+1] += ey*rz*x_len
                        s[ix, iy+1, iz+1] += ry*rz*x_len
                    if xyz == 1:
                        s[ix, iy, iz] += ex*ez*x_len
                        s[ix+1, iy, iz] += rx*ez*x_len
                        s[ix, iy, iz+1] += ex*rz*x_len
                        s[ix+1, iy, iz+1] += rx*rz*x_len
                    if xyz == 2:
                        s[ix, iy, iz] += ex*ey*x_len
                        s[ix+1, iy, iz] += rx*ey*x_len
                        s[ix, iy+1, iz] += ex*ry*x_len
                        s[ix+1, iy+1, iz] += rx*ry*x_len

    # Ensure unity (should not be necessary).
    sum_s = abs(s.sum())
    if abs(sum_s-1) > 1e-6:
        # Print is always shown and simpler, warn for the CLI logs.
        msg = f"Normalizing Source: {sum_s:.10f}."
        print(f"* WARNING :: {msg}")
        warnings.warn(msg, UserWarning)
        s /= sum_s


def _rotation(azm, dip):
    """Rotation factors for RHS with positive z upwards.

    Easting is x, Northing is y, and positive upwards is z. All functions
    should use this rotation to ensure they use all the same definition.

    Parameters
    ----------
    azm : float
        Azimuth (°): horizontal deviation from x-axis, anti-clockwise.

    dip: float
        Dip (°): vertical deviation from xy-plane up-wards.


    Returns
    -------
    rot : ndarray (3,)
        Rotation factors (x, y, z).

    """
    return np.array([cosdg(azm)*cosdg(dip), sindg(azm)*cosdg(dip), sindg(dip)])


def _finite_dipole_from_point_dipole(src, length):
    """Return finite dipole of length given a point dipole."""
    factors = _rotation(*src[3:])*length/2
    return np.ravel(src[:3] + np.stack([-factors, factors]), 'F')


def _square_loop_from_point_dipole(src, length):
    """Return points of a square loop of length x length m perp. to dipole."""
    half_diagonal = np.sqrt(2)*length/2
    rot_hor = _rotation(src[3]+90, 0)*half_diagonal
    rot_ver = _rotation(src[3], src[4]+90)*half_diagonal
    points = src[:3]+np.stack([rot_hor, rot_ver, -rot_hor, -rot_ver, rot_hor])
    return points.T
